import django_tables2 as tables

from app.models import Foo
from crud_views.lib.table import Table, LinkChildColumn, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet

cv_foo = ViewSet(
    model=Foo,
    name="foo",
)


class FooTable(Table):
    id = LinkDetailColumn()
    name = tables.Column()
    bar = LinkChildColumn(name="bar", verbose_name="Bar", empty_values=())


class FooListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Foo
    table_class = FooTable
    cv_viewset = cv_foo
    cv_list_actions = ["detail", "update", "delete"]


class FooDetailView(DetailViewPermissionRequired):
    model = Foo
    cv_viewset = cv_foo
    cv_property_display = [
        {
            "title": "Properties",
            "properties": ["id", "name"],
        },
    ]


class FooUpdateView(UpdateViewPermissionRequired):
    model = Foo
    fields = ["name"]
    cv_viewset = cv_foo


class FooCreateView(CreateViewPermissionRequired):
    model = Foo
    fields = ["name"]
    cv_viewset = cv_foo


class FooDeleteView(DeleteViewPermissionRequired):
    model = Foo
    cv_viewset = cv_foo



import django_filters
import django_tables2 as tables
from crispy_forms.layout import Layout, Row
from django.forms.fields import CharField
from django.utils.translation import gettext as _

from app.models import Author
from crud_views.lib.crispy import Column4, CrispyModelForm, CrispyModelViewMixin, CrispyDeleteForm, Column12
from crud_views.lib.table import Table, LinkChildColumn, UUIDLinkDetailColumn
from crud_views.lib.table.columns import NaturalTimeColumn, NaturalDayColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    MessageMixin,
    ListViewTableMixin,
    ListViewTableFilterMixin,
    ListViewPermissionRequired,
    OrderedUpViewPermissionRequired,
    OrderedUpDownPermissionRequired,
    DeleteViewPermissionRequired,
    RedirectChildView,
)
from crud_views.lib.views.form import CustomFormViewPermissionRequired
from crud_views.lib.views.list import ListViewFilterFormHelper
from crud_views.lib.viewset import ViewSet

cv_author = ViewSet(model=Author, name="author", icon_header="fa-regular fa-user")


class AuthorCreateForm(CrispyModelForm):
    submit_label = _("Create")

    class Meta:
        model = Author
        fields = ["first_name", "last_name", "pseudonym"]

    def get_layout_fields(self):
        return Row(Column4("first_name"), Column4("last_name"), Column4("pseudonym"))


class AuthorUpdateForm(AuthorCreateForm):
    """
    Update form has the same fields as the create form
    """

    submit_label = _("Update")


class AuthorFilterFormHelper(ListViewFilterFormHelper):
    layout = Layout(
        Row(Column4("first_name"), Column4("last_name")),
    )


class AuthorFilter(django_filters.FilterSet):
    first_name = django_filters.CharFilter(lookup_expr="icontains")
    last_name = django_filters.CharFilter(lookup_expr="icontains")

    class Meta:
        model = Author
        fields = [
            "first_name",
            "last_name",
        ]


class AuthorTable(Table):
    id = UUIDLinkDetailColumn(attrs=Table.ca.ID)
    first_name = tables.Column()
    last_name = tables.Column()
    pseudonym = tables.Column(attrs=Table.ca.w20)
    books = LinkChildColumn(name="book", verbose_name="Books", attrs=Table.ca.w10)
    created_dt = NaturalDayColumn()
    modified_dt = NaturalTimeColumn()


class AuthorListView(ListViewTableMixin, ListViewTableFilterMixin, ListViewPermissionRequired):
    # table config
    table_class = AuthorTable

    # filter config
    filterset_class = AuthorFilter
    formhelper_class = AuthorFilterFormHelper

    # viewset config
    cv_viewset = cv_author
    cv_list_actions = ["detail", "update", "delete", "up", "down", "redirect_child", "contact"]


class AuthorCreateView(CrispyModelViewMixin, MessageMixin, CreateViewPermissionRequired):
    form_class = AuthorCreateForm
    cv_viewset = cv_author
    cv_message = "Created author »{object}«"


class AuthorUpdateView(CrispyModelViewMixin, MessageMixin, UpdateViewPermissionRequired):
    form_class = AuthorUpdateForm
    cv_viewset = cv_author
    cv_message = "Updated author »{object}«"


class AuthorDeleteView(CrispyModelViewMixin, MessageMixin, DeleteViewPermissionRequired):
    form_class = CrispyDeleteForm
    cv_viewset = cv_author
    cv_message = "Deleted author »{object}«"


class AuthorDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_author

    cv_property_display = [
        {
            "title": _("Attributes"),
            "icon": "tag",
            "description": _("Core author information"),
            "properties": [
                "xyz",
                {"path": "full_name", "detail": _("Computed from first and last name")},
                "first_name",
                "last_name",
            ],
        },
        {
            "title": _("Extra"),
            "icon": "circle-info",
            "description": _("Additional metadata and computed values"),
            "properties": [
                {"path": "id", "title": "UUID", "detail": _("Unique identifier")},
                "pseudonym",
                {"path": "book_count", "detail": _("Total number of books by this author")},
                "abc",
            ],
        },
    ]


class AuthorUpView(MessageMixin, OrderedUpViewPermissionRequired):
    cv_viewset = cv_author
    cv_message = "Successfully moved author »{object}« up"


class AuthorDownView(MessageMixin, OrderedUpDownPermissionRequired):
    cv_viewset = cv_author
    cv_message = "Successfully moved author »{object}« down"


class RedirectBooksView(RedirectChildView):
    cv_action_label = _("Goto Books")
    cv_redirect = "book"
    cv_redirect_key = "list"
    cv_icon_action = "fa-regular fa-address-book"

    cv_viewset = cv_author


class AuthorContactForm(CrispyModelForm):
    submit_label = _("Send")

    subject = CharField(label="Subject", required=True)
    body = CharField(label="Body", required=True)

    class Meta:
        model = Author
        fields = ["subject", "body"]

    def get_layout_fields(self):
        return Column12("subject"), Column12("body")


class AuthorContactView(MessageMixin, CrispyModelViewMixin, CustomFormViewPermissionRequired):
    cv_key = "contact"
    cv_path = "contact"
    cv_icon_action = "fa-solid fa-envelope"
    cv_viewset = cv_author
    form_class = AuthorContactForm

    cv_message_template_code = "Successfully contacted author »{object}«"

    cv_context_actions = ["parent", "detail", "update", "delete", "contact"]
    cv_header_template_code = _("Contact Author")
    cv_paragraph_template_code = _("Send a message to the Author")
    cv_action_label_template_code = _("Contact Author")
    cv_action_short_label_template_code = _("Contact Author Short")

    def form_valid(self, form):
        return super().form_valid(form)



import re
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from crud_views.lib.viewset import ViewSet

from pydantic import BaseModel, field_validator

REG_PK_NAME = re.compile(r"[a-z_]+_pk")
REG_ATTRIBUTE = re.compile(r"[a-z][a-z_]*")


class ParentViewSet(BaseModel):
    name: str  # the name of the parent ViewSet
    attribute: str | None = None  # the model attribute with the ForeignKey to the parent model
    pk_name: str | None = (
        None  # the name of the primary key in the url pattern with _pk suffix, defaults to parent ViewSet name plus "_pk"
    )
    many_to_many_through_attribute: str | None = (
        None  # the name of the parent's manay to many attribute that points to the child
    )

    @field_validator("attribute", mode="plain")  # noqa
    @classmethod
    def validate_attribute(cls, value: Any) -> Any:
        """
        attribute must be lowercase alpha with underscores
        """
        if value is None:
            return value
        if not REG_ATTRIBUTE.match(value):
            raise ValueError(f"attribute must be lowercase alpha with underscores at {cls}")
        return value

    @field_validator("pk_name", mode="plain")  # noqa
    @classmethod
    def validate_pk_name(cls, value: Any) -> Any:
        """
        pk_name must be lowercase alpha with underscores ending with _p
        """
        if value is None:
            return value
        if not REG_PK_NAME.match(value):
            raise ValueError(f"pk_name must be lowercase alpha with underscores ending with _pk at {cls}")
        return value

    @property
    def viewset(self) -> "ViewSet":
        """
        Get parent ViewSet from registry
        """
        from crud_views.lib.viewset import ViewSet

        return ViewSet.get_viewset(self.name)

    def get_pk_name(self) -> str:
        """
        The make of the parent's primary key in the url pattern
        """
        return self.pk_name if self.pk_name else f"{self.viewset.name}_pk"

    def get_attribute(self) -> str:
        """
        The model attribute with the ForeignKey to the parent model, defaults to the parent's ViewSet name
        """
        return self.attribute if self.attribute else self.viewset.name



from crispy_forms.layout import Row
from django.forms import modelform_factory

from app.models.poly import PolyOne
from crud_views.lib.crispy import Column6
from crud_views.lib.crispy.form import CrispyForm

_PolyOneForm = modelform_factory(
    PolyOne,
    fields=[
        "shared",
        "one",
    ],
)


class PolyOneForm(CrispyForm, _PolyOneForm):
    def get_layout_fields(self):
        return (Row(Column6("shared"), Column6("one")),)



import uuid

from django.db import models
from ordered_model.models import OrderedModel
from polymorphic.models import PolymorphicModel


class Author(OrderedModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    first_name = models.CharField(max_length=100, verbose_name="Vorname")
    last_name = models.CharField(max_length=100)
    pseudonym = models.CharField(max_length=100, blank=True, null=True)
    created_dt = models.DateTimeField(auto_now_add=True)
    modified_dt = models.DateTimeField(auto_now=True)

    class Meta(OrderedModel.Meta):
        verbose_name = "Autor"
        verbose_name_plural = "Autoren"

    def __str__(self):
        return f"{self.first_name} {self.last_name}"

    @property
    def xyz(self):
        return "xyz-prop"

    @property
    def abc(self):
        return True


class Book(OrderedModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    title = models.CharField(max_length=100)
    price = models.DecimalField(max_digits=10, decimal_places=2)
    author = models.ForeignKey(Author, on_delete=models.CASCADE)
    created_dt = models.DateTimeField(auto_now_add=True)
    modified_dt = models.DateTimeField(auto_now=True)

    def __str__(self):
        return f"{self.title} by {self.author}"


class Foo(models.Model):
    name = models.CharField(max_length=100)

    def __str__(self):
        return f"{self.name}"


class Bar(models.Model):
    foo = models.ForeignKey(Foo, on_delete=models.CASCADE)
    name = models.CharField(max_length=100)

    def __str__(self):
        return f"{self.name}"


class Baz(models.Model):
    bar = models.ForeignKey(Bar, on_delete=models.CASCADE)
    name = models.CharField(max_length=100)

    def __str__(self):
        return f"{self.name}"


class Poly(PolymorphicModel, models.Model):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    shared = models.CharField(max_length=100)


class PolyOne(Poly):
    one = models.CharField(max_length=100)


class PolyTwo(Poly):
    two = models.CharField(max_length=100)


class Detail(OrderedModel):
    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    integer = models.IntegerField(verbose_name="An integer field")
    number = models.FloatField(verbose_name="A float field")
    char = models.CharField(max_length=100, verbose_name="Text field")
    text = models.TextField(verbose_name="Multiline Text")
    boolean = models.BooleanField(null=True, default=None, verbose_name="A boolean value")
    boolean_two = models.BooleanField(null=True, default=None, verbose_name="Another boolean value")
    date = models.DateField(verbose_name="A date field")
    date_time = models.DateTimeField(verbose_name="A date field with time")
    author = models.ForeignKey(
        Author, verbose_name="A foreign key field", on_delete=models.SET_NULL, blank=True, null=True
    )
    foo = models.ManyToManyField(Foo, verbose_name="Foo selected")
    created_dt = models.DateTimeField(auto_now_add=True)
    modified_dt = models.DateTimeField(auto_now=True)

    def __str__(self):
        return f"{self.id}"

    @property
    def model_prop(self):
        return True


#########################


class Person(models.Model):
    name = models.CharField(max_length=128)

    def __str__(self):
        return self.name


class Group(models.Model):
    name = models.CharField(max_length=128)
    members = models.ManyToManyField(Person, through="Membership", related_name="group")

    def __str__(self):
        return self.name


class Membership(models.Model):
    person = models.ForeignKey(Person, on_delete=models.CASCADE)
    group = models.ForeignKey(Group, on_delete=models.CASCADE)
    date_joined = models.DateField()
    invite_reason = models.CharField(max_length=64)

    class Meta:
        constraints = [models.UniqueConstraint(fields=["person", "group"], name="unique_person_group")]



from crud_views.lib.view import CrudViewPermissionRequiredMixin
from .action import ActionView


class OrderedCheckBase:
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        from ordered_model.models import OrderedModel

        # todo: move to system checks
        if not issubclass(self.model, OrderedModel):
            raise ValueError(f"{self.model} is not a subclass of OrderedModel")


class OrderedUpView(ActionView):
    cv_key = "up"
    cv_path = "up"
    cv_backend_only = True

    # texts and labels
    cv_action_label_template: str | None = "crud_views/snippets/action/up.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/up.html"

    # icons
    cv_icon_action = "fa-regular fa-circle-up"

    # messages
    cv_message_template: str | None = "crud_views/snippets/message/up.html"

    def action(self, context: dict) -> bool:
        return self.up(context)

    def up(self, context: dict) -> bool:
        self.object.up()
        self.object.save()
        return True


class OrderedUpViewPermissionRequired(CrudViewPermissionRequiredMixin, OrderedUpView):  # this file
    cv_permission = "change"


class OrderedDownView(ActionView):
    cv_key = "down"
    cv_path = "down"
    cv_backend_only = True

    # texts and labels
    cv_action_label_template: str | None = "crud_views/snippets/action/down.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/down.html"

    # icons
    cv_icon_action = "fa-regular fa-circle-down"

    # messages
    cv_message_template: str | None = "crud_views/snippets/message/down.html"

    def action(self, context: dict) -> bool:
        return self.down(context)

    def down(self, context: dict) -> bool:
        self.object.down()
        self.object.save()
        return True


class OrderedUpDownPermissionRequired(CrudViewPermissionRequiredMixin, OrderedDownView):  # this file
    cv_permission = "change"



from collections import OrderedDict
from datetime import date
from typing import List, Dict, Type

import django_tables2 as tables
from crispy_forms.layout import Row, LayoutObject
from django.contrib.contenttypes.models import ContentType
from django.db.models import Model
from django.forms.models import inlineformset_factory
from django.utils.translation import gettext_lazy as _

from app.models.poly import Poly, PolyOne, PolyTwo, PolyTwoChoice, PolyThree, PolyAnswerText, PolyAnswerNumber
from crud_views.lib.crispy import Column4, CrispyModelViewMixin, Column8
from crud_views.lib.crispy.form import CrispyForm, CrispyModelForm
from crud_views.lib.formsets import FormSet, FormSets, InlineFormSet
from crud_views.lib.formsets.mixins import PolymorphicFormSetMixin
from crud_views_polymorphic.lib import (
    PolymorphicCreateSelectView,
    PolymorphicCreateViewPermissionRequired,
    PolymorphicUpdateViewPermissionRequired,
    PolymorphicDetailViewPermissionRequired,
)
from crud_views_polymorphic.lib.create_select import PolymorphicContentTypeForm
from crud_views_polymorphic.lib.delete import PolymorphicDeleteViewPermissionRequired
from crud_views.lib.table import Table, UUIDLinkDetailColumn
from crud_views.lib.views import ListViewPermissionRequired, ListViewTableMixin, CreateViewParentMixin
from crud_views.lib.viewset import ViewSet, ParentViewSet
from .answer import PolyAnswerFormSet
from .one import PolyOneForm
from .three import PolyThreeForm
from .two import PolyTwoForm

cv_poly_formset = ViewSet(
    model=Poly,
    name="poly",
    icon_header="fa-solid fa-sun",
    parent=ParentViewSet(
        name="parent",
        many_to_many_through_attribute="polys",
        attribute="parents",  # related_name of many2many field
    ),
)


class PolyTable(Table):
    id = UUIDLinkDetailColumn(attrs=Table.ca.ID)
    shared = tables.Column(attrs=Table.ca.w80)


class PolyListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Poly
    table_class = PolyTable
    cv_viewset = cv_poly_formset
    cv_list_actions = ["detail", "update", "delete"]
    cv_context_actions = [
        "parent",
        "create_select",
    ]


class ChoiceForm(CrispyModelForm):
    class Meta:
        model = PolyTwoChoice
        fields = [
            "choice",
        ]


class ChoiceInlineFormSet(InlineFormSet):
    def get_helper_layout_fields(self) -> List[LayoutObject]:
        return [Row(Column8("choice"), self.form_control_col4)]


ChoiceFormSet = inlineformset_factory(
    PolyTwo,
    PolyTwoChoice,
    formset=ChoiceInlineFormSet,
    form=ChoiceForm,
    fields=[
        "choice",
    ],
    extra=1,
    # this adds fields to the form
    can_delete=True,
    can_delete_extra=True,
    can_order=True,
)

formset_answers = FormSet(  # noqa
    title="Answers",
    klass=PolyAnswerFormSet,
    fields=["choice", "help_text"],
    pk_field="id",
)

cv_formsets_one_update: FormSets = FormSets(
    scripts=True,  # we have no controls, therefore no scripts
    formsets=OrderedDict(answers=formset_answers),
)

formset_choices = FormSet(  # noqa
    title="Choices",
    klass=ChoiceFormSet,
    fields=[
        "choice",
    ],
    pk_field="id",
    pk=FormSet.PK.UUID,
)

cv_formsets_two_create: FormSets = FormSets(
    scripts=True,  # we have no controls, therefore no scripts
    formsets=OrderedDict(choices=formset_choices),
)

cv_formsets_two_update: FormSets = FormSets(
    scripts=True,  # we have no controls, therefore no scripts
    formsets=OrderedDict(
        choices=formset_choices,
        answers=formset_answers,
    ),
)


class CrispyPolymorphicContentTypeForm(CrispyForm, PolymorphicContentTypeForm):
    submit_label = _("Select")

    def get_layout_fields(self):
        return Row(Column4("polymorphic_ctype_id"))


class PolyCreateSelectView(CrispyModelViewMixin, PolymorphicCreateSelectView):
    model = Poly
    form_class = CrispyPolymorphicContentTypeForm
    cv_viewset = cv_poly_formset


class PolyCreateView(
    CrispyModelViewMixin, CreateViewParentMixin, PolymorphicFormSetMixin, PolymorphicCreateViewPermissionRequired
):
    model = Poly
    cv_viewset = cv_poly_formset
    cv_context_actions = ["home"]
    cv_formsets_required: bool = False
    cv_polymorphic_formsets: Dict[Type[Model], FormSets] = {
        PolyOne: None,
        PolyTwo: cv_formsets_two_create,
        PolyThree: None,  # this model has explicitly no formsets
    }
    polymorphic_forms = {PolyOne: PolyOneForm, PolyTwo: PolyTwoForm, PolyThree: PolyThreeForm}

    def cv_parent_many_to_many_through_defaults(self, instance, parent_instance, m2m) -> dict:
        defaults = super().cv_parent_many_to_many_through_defaults(instance, parent_instance, m2m)
        defaults.update({"date_joined": date.today()})
        return defaults

    def cv_form_valid_hook(self, context: dict):
        ct = ContentType.objects.get(id=context["polymorphic_ctype_id"])
        if ct.model_class() in [PolyOne, PolyTwo]:
            PolyAnswerText.objects.create(poly=self.object)
            PolyAnswerNumber.objects.create(poly=self.object)


class PolyUpdateView(CrispyModelViewMixin, PolymorphicFormSetMixin, PolymorphicUpdateViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly_formset
    cv_formsets_required: bool = False
    cv_polymorphic_formsets: Dict[Type[Model], FormSets] = {
        PolyOne: cv_formsets_one_update,
        PolyTwo: cv_formsets_two_update,
        PolyThree: None,  # this model has explicitly no formsets
    }
    polymorphic_forms = {PolyOne: PolyOneForm, PolyTwo: PolyTwoForm, PolyThree: PolyThreeForm}


class PolyDeleteView(PolymorphicDeleteViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly_formset


class PolyDetailView(PolymorphicDetailViewPermissionRequired):
    model = Poly
    cv_viewset = cv_poly_formset
    cv_property_display = [
        {
            "title": "Properties",
            "icon": "shapes",
            "description": "Shared polymorphic fields",
            "properties": ["id", "shared"],
        },
    ]



import pytest
from django.test.client import Client

from crud_views.lib.viewset import ViewSet
from tests.lib.helper.boostrap5 import Table
from tests.test1.app.models import Book


@pytest.mark.django_db
def test_child_viewset_has_parent(cv_book: ViewSet):
    assert cv_book.parent is not None
    assert cv_book.has_parent is True
    assert cv_book.parent.name == "publisher"


@pytest.mark.django_db
def test_child_viewset_url_patterns(cv_book: ViewSet):
    patterns = cv_book.urlpatterns
    assert len(patterns) >= 5

    # All routes should contain the parent prefix
    routes = [p.pattern.regex.pattern for p in patterns]
    for route in routes:
        assert "publisher/" in route, f"Expected 'publisher/' in route: {route}"
        assert r"\d+" in route, f"Expected INT pk regex in route: {route}"


@pytest.mark.django_db
def test_child_viewset_router_names(cv_book: ViewSet):
    assert cv_book.get_router_name("list") == "book-list"
    assert cv_book.get_router_name("detail") == "book-detail"
    assert cv_book.get_router_name("create") == "book-create"
    assert cv_book.get_router_name("update") == "book-update"
    assert cv_book.get_router_name("delete") == "book-delete"


@pytest.mark.django_db
def test_child_list_books(client_user_book_view: Client, cv_book, publisher_penguin, book_hitchhiker):
    url = f"/publisher/{publisher_penguin.pk}/book/"
    response = client_user_book_view.get(url)
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 1
    assert "Hitchhiker" in table.rows[0].columns[1].text


@pytest.mark.django_db
def test_child_list_filters_by_parent(
    client_user_book_view: Client,
    cv_book,
    publisher_penguin,
    publisher_harpercollins,
    book_hitchhiker,
    book_other_publisher,
):
    # List books under Penguin — should only see Hitchhiker
    url = f"/publisher/{publisher_penguin.pk}/book/"
    response = client_user_book_view.get(url)
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 1
    assert "Hitchhiker" in table.rows[0].columns[1].text

    # List books under HarperCollins — should only see Other Book
    url = f"/publisher/{publisher_harpercollins.pk}/book/"
    response = client_user_book_view.get(url)
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 1
    assert "Other Book" in table.rows[0].columns[1].text


@pytest.mark.django_db
def test_child_create_book(client_user_book_add: Client, cv_book, publisher_penguin):
    url = f"/publisher/{publisher_penguin.pk}/book/create/"
    response = client_user_book_add.get(url)
    assert response.status_code == 200

    response = client_user_book_add.post(url, {"title": "1984"})
    assert response.status_code == 302

    book = Book.objects.get(title="1984")
    assert book.publisher == publisher_penguin


@pytest.mark.django_db
def test_child_update_book(client_user_book_change: Client, cv_book, publisher_penguin, book_hitchhiker):
    url = f"/publisher/{publisher_penguin.pk}/book/{book_hitchhiker.pk}/update/"
    response = client_user_book_change.get(url)
    assert response.status_code == 200

    response = client_user_book_change.post(url, {"title": "H2G2"})
    assert response.status_code == 302

    book_hitchhiker.refresh_from_db()
    assert book_hitchhiker.title == "H2G2"


@pytest.mark.django_db
def test_child_delete_book(client_user_book_delete: Client, cv_book, publisher_penguin, book_hitchhiker):
    pk = book_hitchhiker.pk
    url = f"/publisher/{publisher_penguin.pk}/book/{pk}/delete/"

    response = client_user_book_delete.get(url)
    assert response.status_code == 200

    response = client_user_book_delete.post(url, {"confirm": True})
    assert response.status_code == 302

    assert not Book.objects.filter(pk=pk).exists()


@pytest.mark.django_db
def test_child_detail_book(client_user_book_view: Client, cv_book, publisher_penguin, book_hitchhiker):
    url = f"/publisher/{publisher_penguin.pk}/book/{book_hitchhiker.pk}/detail/"
    response = client_user_book_view.get(url)
    assert response.status_code == 200

    content = response.content.decode("utf-8")
    assert "Hitchhiker" in content


@pytest.mark.django_db
def test_child_permissions_anonymous(client: Client, cv_book, publisher_penguin, book_hitchhiker):
    """Anonymous user should be redirected to login."""
    url = f"/publisher/{publisher_penguin.pk}/book/"
    response = client.get(url)
    assert response.status_code == 302
    assert "/accounts/login/" in response.url


@pytest.mark.django_db
def test_child_permissions_no_perm(client_user_a: Client, cv_book, publisher_penguin, book_hitchhiker):
    """Authenticated user without permissions should get 403."""
    url = f"/publisher/{publisher_penguin.pk}/book/"
    response = client_user_a.get(url)
    assert response.status_code == 403



from django.core.management import BaseCommand

from app.models import Person


class Command(BaseCommand):
    help = "Show Django CRUD views"

    def handle(self, *args, **options):

        person = Person.objects.filter(group__id=1).first()

        person.membership_set.all()



import pytest
from django.contrib.auth import get_user_model
from django.test.client import Client

from tests.lib.helper.boostrap5 import Table

User = get_user_model()


@pytest.mark.django_db
def test_debug_view(user_author_view: User, cv_author, author_douglas_adams, author_terry_pratchett):
    client = Client()

    client.force_login(user_author_view)
    response = client.get("/author/")
    assert response.status_code == 200

    lst = Table(response)
    headers = lst.headers

    for header in headers:
        print(header.text, header.orderable)
    for row in lst.rows:
        for action in row.actions:
            print(action.title, action.href)
        # for col in row.columns:
        #    col.text

        # if header.is_action:
        #    for action in header.actions:
        #        print(action.text, action.href)


@pytest.mark.skip(reason="disabled")
def test_fail():
    assert False, "this is a test"



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("cvw", "0001_initial"),
    ]

    operations = [
        migrations.RenameField(
            model_name="workflowinfo",
            old_name="workflow_object_id",
            new_name="workflow_object_pk",
        ),
        migrations.AlterField(
            model_name="workflowinfo",
            name="workflow_object_pk",
            field=models.CharField(max_length=255),
        ),
        migrations.AddIndex(
            model_name="workflowinfo",
            index=models.Index(
                fields=["workflow_object_pk", "workflow_object_content_type"],
                name="cvw_workflo_workflo_idx",
            ),
        ),
    ]



import pytest
from django.contrib.auth.models import User
from django.contrib.messages import get_messages
from django.test.client import Client

from tests.lib.helper.user import user_viewset_permission


@pytest.fixture
def client_publisher_add_view(client, cv_publisher) -> Client:
    """Client with both add and view permissions (needed to follow redirect to list)."""
    user = User.objects.create_user(username="user_pub_add_view", password="password")
    user_viewset_permission(user, cv_publisher, "add")
    user_viewset_permission(user, cv_publisher, "view")
    client.force_login(user)
    return client


@pytest.fixture
def client_publisher_change_view(client, cv_publisher) -> Client:
    """Client with both change and view permissions."""
    user = User.objects.create_user(username="user_pub_change_view", password="password")
    user_viewset_permission(user, cv_publisher, "change")
    user_viewset_permission(user, cv_publisher, "view")
    client.force_login(user)
    return client


@pytest.fixture
def client_publisher_delete_view(client, cv_publisher) -> Client:
    """Client with both delete and view permissions."""
    user = User.objects.create_user(username="user_pub_delete_view", password="password")
    user_viewset_permission(user, cv_publisher, "delete")
    user_viewset_permission(user, cv_publisher, "view")
    client.force_login(user)
    return client


@pytest.mark.django_db
def test_create_message(client_publisher_add_view: Client, cv_publisher):
    response = client_publisher_add_view.post("/publisher/create/", {"name": "Penguin"}, follow=True)
    assert response.status_code == 200

    messages = list(get_messages(response.wsgi_request))
    assert len(messages) == 1
    assert "Created" in str(messages[0])
    assert "Penguin" in str(messages[0])


@pytest.mark.django_db
def test_update_message(client_publisher_change_view: Client, cv_publisher, publisher_penguin):
    pk = publisher_penguin.pk
    response = client_publisher_change_view.post(f"/publisher/{pk}/update/", {"name": "Penguin Classics"}, follow=True)
    assert response.status_code == 200

    messages = list(get_messages(response.wsgi_request))
    assert len(messages) == 1
    assert "Updated" in str(messages[0])
    assert "Penguin Classics" in str(messages[0])


@pytest.mark.django_db
def test_delete_message(client_publisher_delete_view: Client, cv_publisher, publisher_penguin):
    pk = publisher_penguin.pk
    response = client_publisher_delete_view.post(f"/publisher/{pk}/delete/", {"confirm": True}, follow=True)
    assert response.status_code == 200

    messages = list(get_messages(response.wsgi_request))
    assert len(messages) == 1
    assert "Deleted" in str(messages[0])
    assert "Penguin" in str(messages[0])


@pytest.mark.django_db
def test_no_message_on_get(client_user_publisher_view: Client, cv_publisher, publisher_penguin):
    """GET requests should not produce any flash messages."""
    pk = publisher_penguin.pk
    response = client_user_publisher_view.get(f"/publisher/{pk}/detail/")
    assert response.status_code == 200

    messages = list(get_messages(response.wsgi_request))
    assert len(messages) == 0



from django.apps import AppConfig


class CrudViewsPlainConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "crud_views_plain"

    def ready(self):
        pass
        # todo: run checks
        # import viewset.checks  # noqa



from typing import Iterable

from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import SuspiciousOperation
from django.db import transaction
from django.utils.translation import gettext_lazy as _

from crud_views.lib.check import Check, CheckAttribute, CheckExpression
from crud_views.lib.view import CrudViewPermissionRequiredMixin
from crud_views.lib.views.form import CustomFormView
from ..models import WorkflowInfo
from .mixins import WorkflowModelMixin


class WorkflowView(CustomFormView):
    """
    Base view for workflow transitions using the WorkflowForm
    """

    form_class = None
    cv_key = "workflow"
    cv_path = "workflow"

    template_name = "crud_views_workflow/view_workflow.html"

    cv_icon_header = "fa-solid fa-diagram-project"
    cv_icon_action = "fa-solid fa-diagram-project"
    cv_message_template_code = _("Successfully processed workflow step on »{{ object }}«")
    cv_header_template_code = _("Process workflow")
    cv_paragraph_template_code = _("Process workflow step")
    cv_action_label_template_code = _("Process workflow")
    cv_action_short_label_template_code = _("Process workflow")

    cv_transition_label = _("Select a possible workflow action to take")
    cv_transition_help_text = None
    cv_comment_label = _("Please provide a comment for your workflow step")
    cv_comment_help_text = None

    @classmethod
    def checks(cls) -> Iterable[Check]:
        yield from super().checks()
        # form_class must be set to a WorkflowForm subclass
        yield CheckAttribute(context=cls, id="E230", attribute="form_class")
        # workflow-specific label attributes must not be None
        yield CheckAttribute(context=cls, id="E231", attribute="cv_transition_label")
        yield CheckAttribute(context=cls, id="E232", attribute="cv_comment_label")
        # the model associated with this view must extend WorkflowModelMixin
        if cls.cv_viewset is not None:
            model = cls.cv_viewset.model
            yield CheckExpression(
                context=cls,
                id="E233",
                expression=issubclass(model, WorkflowModelMixin),
                msg=f"Model »{model.__name__}» must extend WorkflowModelMixin",
            )
            # required WorkflowModelMixin class attributes must be set on the model
            if issubclass(model, WorkflowModelMixin):
                yield CheckAttribute(context=model, id="E234", attribute="STATE_CHOICES")
                yield CheckAttribute(context=model, id="E235", attribute="STATE_BADGES")

    def get_form(self, form_class=None):
        form = super().get_form(form_class)

        # apply custom labels and help texts
        form.fields["transition"].label = self.cv_transition_label
        form.fields["transition"].help_text = self.cv_transition_help_text
        form.fields["comment"].label = self.cv_comment_label
        form.fields["comment"].help_text = self.cv_comment_help_text

        return form

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()
        kwargs.update(self.object.workflow_get_form_kwargs(self.request.user))
        return kwargs

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["has_workflow_choices"] = self.object.workflow_has_any_possible_transition(self.request.user)
        return context

    # todo: cv_form_valid instead of cv_form_valid_hook?
    def cv_form_valid_hook(self, context: dict):
        """
        Process workflow transition
        """
        form = context["form"]
        transition = form.cleaned_data["transition"]
        comment = form.cleaned_data["comment"]
        if isinstance(comment, str):
            comment = comment.strip()
        if not comment:
            comment = None

        # some security: make sure the user is allowed to take this transition
        if not self.object.workflow_has_transition(self.request.user, transition):
            raise SuspiciousOperation(f"Invalid transition {transition}")

        # get method to call
        wf_method = getattr(self.object, transition, None)
        assert callable(wf_method), f"Invalid transition {transition}"

        # todo: configurable
        with transaction.atomic():
            # get old state
            state_old = self.object.state

            # execute the workflow
            data = wf_method(by=self.request.user, comment=comment)

            # save object
            self.object.save()

            # new state
            state_new = self.object.state
            user = self.request.user

            # and save the workflow state
            info = WorkflowInfo.objects.create(
                transition=transition,
                state_old=state_old,
                state_new=state_new,
                comment=comment,
                user=user,
                data=data,
                workflow_object_pk=str(self.object.pk),
                workflow_object_content_type=ContentType.objects.get_for_model(self.object),
            )

            # call hook
            self.on_transition(info, transition, state_old, state_new, comment, user, data)

    def on_transition(self, info, transition, state_old, state_new, comment, user, data):
        """
        Override this to do additional stuff after a transition
        """
        pass


class WorkflowViewPermissionRequired(CrudViewPermissionRequiredMixin, WorkflowView):
    cv_permission = "change"



from django.apps import AppConfig


class CrudViewsConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "crud_views"

    def ready(self):
        import crud_views.checks  # noqa



from __future__ import annotations

from functools import cached_property
from typing import Dict, List, Type, Any, Iterable, Tuple, TYPE_CHECKING

if TYPE_CHECKING:
    from crud_views.lib.viewset import ViewSet

from crud_views.lib import check
from crud_views.lib.check import Check, CheckAttributeReg, CheckAttribute, CheckTemplateOrCode
from crud_views.lib.exceptions import cv_raise, ParentViewSetError, CrudViewError
from django.contrib.auth import get_user_model
from django.contrib.auth.mixins import PermissionRequiredMixin
from django.db.models import Model
from django.shortcuts import get_object_or_404
from django.template import Context as TemplateContext, Template
from django.template.loader import render_to_string
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _
from typing_extensions import Self

from .buttons import ContextButton
from .context import ViewContext
from .meta import CrudViewMetaClass
from ..settings import crud_views_settings

User = get_user_model()


class CrudView(metaclass=CrudViewMetaClass):
    """
    A view that is part of a ViewSet
    """

    cv_viewset: "ViewSet" = None
    cv_key: str = None  # the key to register the view (i.e. detail, list, create, update, delete)
    cv_path: str = None  # i.e. detail, update or "" for list views
    cv_object: bool = True  # view has object context (only list views do not have object context)
    cv_backend_only: bool = (
        False  # views is only available in the backend, so i.e. title and paragraph templates are not required
    )
    cv_list_actions: List[str] | None = None  # actions for the list view
    cv_list_action_method: str = "get"  # method to call for list actions
    cv_context_actions: List[str] | None = None  # context actions for the view (top right)
    cv_home_key: str | None = "list"  # home url, defaults to list
    cv_success_key: str | None = "list"  # success url, defaults to list
    cv_cancel_key: str | None = "list"  # cancel url, defaults to list
    cv_parent_key: str | None = "list"  # parent key, defaults to list todo: does this make sense at all?

    cv_extends_template: str | None = None  # template to extend

    # texts and labels
    cv_header_template: str | None = None  # template snippet to render header label
    cv_header_template_code: str | None = None  # template code to render header label
    cv_paragraph_template: str | None = None  # template snippet to render paragraph
    cv_paragraph_template_code: str | None = None  # template code to render paragraph
    cv_action_label_template: str | None = None  # template snippet to render action label
    cv_action_label_template_code: str | None = None  # template code to render action label
    cv_action_short_label_template: str | None = None  # template snippet to render short action label without icons
    cv_action_short_label_template_code: str | None = None  # template code to render short  action label  without icons
    cv_filter_header_template: str | None = None  # template snippet to render filter header
    cv_filter_header_template_code: str | None = None  # template code to render filter header
    cv_message_template: str | None = None  # template snippet to render messages
    cv_message_template_code: str | None = None  # template code to render messages

    # icons
    cv_icon_action: str | None = None  # font awesome icon
    cv_icon_header: str | None = None  # font awesome icon

    @classmethod
    def checks(cls) -> Iterable[Check]:
        """
        Iterator of checks for the view
        """
        yield CheckAttributeReg(context=cls, id="E200", attribute="cv_key", **check.REGS["name"])
        yield CheckAttributeReg(context=cls, id="E201", attribute="cv_path", **check.REGS["path"])

        # todo: cv_extends test template if defined

        # todo: reactivate
        # yield ContextActionCheck(context=cls, id="E203", msg="action not defined")

        # templates are required for frontend views
        is_frontend = not cls.cv_backend_only
        if is_frontend:
            for attribute in [
                "cv_header_template",
                "cv_paragraph_template",
                "cv_action_label_template",
                "cv_action_short_label_template",
            ]:
                yield CheckTemplateOrCode(context=cls, attribute=attribute)

    def get_success_url(self) -> str:
        url = self.cv_get_url(key=self.cv_success_key, obj=getattr(self, "object", None))
        return url

    def get_queryset(self):
        return self.cv_viewset.get_queryset(view=self)

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["cv_extends"] = self.cv_get_extends_template()
        return context

    def cv_get_extends_template(self) -> str:
        cv_extends = self.cv_extends_template
        extends = cv_extends if cv_extends else crud_views_settings.extends
        return extends

    @classmethod
    def cv_has_access(cls, user: User, obj: Model | None = None) -> bool:
        return True

    @classmethod
    def render_snippet(cls, data: dict, template: str = None, template_code: str = None) -> str:
        """
        Either render the template_code or the template
        """
        if template_code:
            template = Template(template_code)
            context = TemplateContext(data)
            result = template.render(context)
        elif template:
            result = render_to_string(template, data)
        else:
            raise CrudViewError(f"no template or template_code provided for {cls}")

        # strip leading and trailing whitespaces and mark it as safe
        return mark_safe(result.strip())

    def cv_get_header_icon(self) -> str:
        view_icon = self.cv_icon_header
        icon = view_icon or self.cv_viewset.icon_header
        return icon

    @property
    def cv_header(self) -> str:
        return self.render_snippet(
            self.cv_get_meta(),
            self.cv_header_template,
            self.cv_header_template_code,
        )

    @property
    def cv_paragraph(self) -> str:
        return self.render_snippet(
            self.cv_get_meta(),
            self.cv_paragraph_template,
            self.cv_paragraph_template_code,
        )

    @classmethod
    def cv_get_action_label(cls, context: ViewContext) -> str:
        return cls.render_snippet(
            cls.cv_viewset.get_meta(context),
            cls.cv_action_label_template,
            cls.cv_action_label_template_code,
        )

    @classmethod
    def cv_get_action_short_label(cls, context: ViewContext) -> str:
        return cls.render_snippet(
            cls.cv_viewset.get_meta(context),
            cls.cv_action_short_label_template,
            cls.cv_action_short_label_template_code,
        )

    @classmethod
    def cv_get_dict(cls, context: ViewContext, **extra) -> Dict[str, Any]:
        """
        Note: This is a classmethod, so the view instance and it's object context is not available here.
              The data this method returns is used to link sibling views.
        """
        data = dict(
            cv_key=cls.cv_key,
            cv_path=cls.cv_path,
            cv_action_label=cls.cv_get_action_label(context=context),
            cv_action_short_label=cls.cv_get_action_short_label(context=context),
            cv_list_actions=cls.cv_list_actions,
            cv_list_action_method=cls.cv_list_action_method,
            cv_context_actions=cls.cv_context_actions,
            cv_home_key=cls.cv_home_key,
            cv_success_key=cls.cv_success_key,
            cv_cancel_key=cls.cv_cancel_key,
            cv_icon_action=cls.cv_icon_action,
            cv_icon_header=cls.cv_icon_header,
        )
        data.update(extra)
        return data

    @classmethod
    def cv_path_contribute(cls) -> str:
        """
        Contribute path to the path of the view
        """
        return ""

    def cv_get_cls(self, key: str | None = None) -> Type[Self]:
        """
        Get the class of the view or for a sibling of the view from ViewSet
        """
        key = key or self.cv_key
        cls = self.__class__ if key == self.cv_key else self.cv_viewset.get_view_class(key)
        return cls

    def cv_get_cls_assert_object(self, key: str | None = None, obj: Model | None = None) -> Type[Self]:
        """
        See cv_get_cls, but assert object context
        """
        cls = self.cv_get_cls(key)
        cv_raise(cls.cv_object is False or cls.cv_object is True and obj, f"view {cls} requires object")
        return cls

    @classmethod
    def cv_get_url_extra_kwargs(cls) -> dict:
        return dict()

    def cv_get_router_and_args(
        self, key: str | None = None, obj=None, extra_kwargs: dict | None = None
    ) -> Tuple[str, tuple, dict]:
        """
        Get the router name, args, kwargs url for a sibling defined by a key
        """
        cls = self.cv_get_cls_assert_object(key, obj)

        if extra_kwargs:
            assert isinstance(extra_kwargs, dict)
        kwargs = extra_kwargs if extra_kwargs else dict()
        args = []

        # if the view requires an object, add pk using the pk_name defined at ViewSet
        if cls.cv_object:
            kwargs[self.cv_viewset.pk_name] = obj.pk
            args.append(obj.pk)

        # get kwargs to pass
        #   1. parent kwargs
        #   2. extra kwargs defined at ViewSet
        #   3. additional kwargs provided by CrudView
        parent_url_args = self.cv_viewset.get_parent_url_args()
        for name in parent_url_args:
            value = self.kwargs.get(name)
            if not value:
                raise ValueError(f"kwarg {name} not found at {self}")
            kwargs[name] = value
            args.append(value)
        kwargs.update(cls.cv_get_url_extra_kwargs())

        args.reverse()
        router_name = self.cv_viewset.get_router_name(key)
        return router_name, tuple(args), kwargs

    def cv_get_url(self, key: str | None = None, obj=None, extra_kwargs: dict | None = None) -> str:
        """
        Get the url for a sibling defined by key
        """
        router_name, args, kwargs = self.cv_get_router_and_args(key=key, obj=obj, extra_kwargs=extra_kwargs)
        url_path = reverse(router_name, kwargs=kwargs)
        return url_path

    def cv_get_view_context(self, **kwargs) -> ViewContext:
        """
        Get the context for the view
        """
        if self.cv_object and "object" not in kwargs:
            kwargs["object"] = self.object

        if "view" not in kwargs:
            kwargs["view"] = self

        return ViewContext(**kwargs)

    def cv_get_context_button(self, key: str) -> ContextButton | None:
        # todo: first look in CrudView context_buttons
        pass

        # then look as ViewSet context_buttons
        for cb in self.cv_viewset.context_buttons:
            if cb.key == key:
                return cb
        return None

    def cv_get_oid(self, key: str, obj: Model | None = None) -> str | None:
        """
        get unique object id
        """
        if not obj:
            return None
        pk = str(obj.pk).replace("-", "").replace(" ", "")
        return f"{self.cv_viewset.name}_{key}_{pk}"

    def cv_get_context(
        self, key: str | None = None, obj: Model | None = None, user: User | None = None, request=None
    ) -> Dict[str, Any]:
        """
        Get template context for this view for a key and an optional object
        """

        # first get the view context
        context = self.cv_get_view_context(object=obj)

        # is the key a context button?
        context_button = self.cv_get_context_button(key)
        if context_button:
            ctx = context_button.get_context(context)
            return ctx

        # the key is a view
        dict_kwargs = dict(
            cv_access=False,
            cv_oid=self.cv_get_oid(key=key, obj=obj),
            cv_url=self.cv_get_url(key=key, obj=obj),
            cv_is_active=self.cv_viewset.get_router_name(key) == context.router_name,
        )

        # get target view class
        cls = self.cv_get_cls_assert_object(key, obj)

        # set up the view context
        context = self.cv_get_view_context(object=obj)

        # check access
        if cls.cv_has_access(user, obj):
            dict_kwargs.update(
                cv_access=True,
            )

        # prepare dict
        data = cls.cv_get_dict(context=context, **dict_kwargs)

        return data

    def get_cancel_button_context(self, obj: Model | None = None, user: User | None = None, request=None) -> dict:
        """
        Get the context for the cancel button
        """
        url = self.cv_get_url(key=self.cv_cancel_key, obj=obj)
        return dict(cv_url=url, cv_action_label=_("Cancel"))

    def cv_get_child_url(self, name: str, key: str, obj: Model | None = None) -> str:
        """
        Get the URL to the child from the current CrudView's context
        """
        viewset = self.cv_viewset.get_viewset(name)
        if viewset.parent.name != self.cv_viewset.name:
            raise ParentViewSetError(f"ViewSet {viewset} is no child of {self.cv_viewset}")
        # todo: check if this is a child of self.cv
        name = viewset.get_router_name(key)
        args = viewset.get_parent_url_args()
        attrs = viewset.get_parent_attributes()
        kw = dict()
        for idx, (arg, attr) in enumerate(zip(args, attrs)):
            if idx == 0 and obj:
                kw[arg] = obj.pk  # it's me, because I'm linking to the child
            else:
                kw[arg] = self.kwargs[arg]  # get value from the view's kwargs
        url = reverse(name, kwargs=kw)
        return url

    def cv_get_meta(self) -> dict:
        """
        Metadata from ViewSet plus ViewContext
        """
        context = self.cv_get_view_context()
        data = self.cv_viewset.get_meta(context=context)

        # add view specific data
        if hasattr(self, "object"):
            data["object"] = self.object

        return data

    def cv_assert_parent(self):
        assert self.cv_viewset.has_parent, f"ViewSet {self.cv_name} has no parent"

    def cv_get_parent_object(self) -> Model:
        """
        Get parent object based on the view's kwargs
        """
        self.cv_assert_parent()

        assert self.cv_viewset.has_parent, "this ViewSet has no parent"

        # get the parent object
        parent_model = self.cv_viewset.get_parent_model()
        arg = self.cv_viewset.get_parent_url_args(first_only=True)
        pk = self.kwargs[arg]  # noqa
        return get_object_or_404(parent_model, pk=pk)

    def cv_get_parent_object_attribute(self) -> str:
        """
        Get the attribute/field that points to the parent object
        """
        self.cv_assert_parent()

        return self.cv_viewset.get_parent_attributes(first_only=True)


# ViewContext uses a string type hint to CrudView, so we need to rebuild the model here
ViewContext.model_rebuild()


class CrudViewPermissionRequiredMixin(PermissionRequiredMixin):
    cv_permission: str = None  # permission required for the view

    @classmethod
    def checks(cls) -> Iterable[Check]:
        """
        Iterator of checks for the view
        """
        yield from super().checks()  # noqa
        # todo
        yield CheckAttribute(context=cls, id="E202", attribute="cv_permission")

    @cached_property
    def permission_required(self) -> str:
        cv_raise(self.cv_permission is not None, f"cv_permission not set at {self}")
        perms = self.cv_viewset.permissions  # noqa
        perm = perms.get(self.cv_permission)
        assert perm, f"permission {self.cv_permission} not found at {self}"
        return perm

    @classmethod
    def cv_has_access(cls, user: User, obj: Model | None = None) -> bool:
        perm = cls.cv_viewset.permissions.get(cls.cv_permission)
        perms = (perm,) if perm else tuple()
        has_access = user.has_perms(perms)
        return has_access



from django.apps import AppConfig


class CrudViewsWorkflowConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "crud_views_workflow"
    label = "cvw"

    def ready(self):
        # todo: check if needed apps are installed: fsm2 fsm2-admin auditlog
        # import crud_views.checks  # noqa
        pass



from crud_views_polymorphic.lib.create import PolymorphicCreateView, PolymorphicCreateViewPermissionRequired
from crud_views_polymorphic.lib.create_select import (
    PolymorphicCreateSelectView,
    PolymorphicCreateSelectViewPermissionRequired,
)
from crud_views_polymorphic.lib.detail import PolymorphicDetailView, PolymorphicDetailViewPermissionRequired
from crud_views_polymorphic.lib.update import PolymorphicUpdateView, PolymorphicUpdateViewPermissionRequired

__all__ = [
    "PolymorphicDetailView",
    "PolymorphicDetailViewPermissionRequired",
    "PolymorphicCreateSelectView",
    "PolymorphicCreateSelectViewPermissionRequired",
    "PolymorphicCreateView",
    "PolymorphicCreateViewPermissionRequired",
    "PolymorphicUpdateView",
    "PolymorphicUpdateViewPermissionRequired",
]



import django_tables2 as tables
from crispy_forms.layout import Row
from django.utils.translation import gettext_lazy as _

from django_object_detail import PropertyConfig

from app.models.campaign import Campaign
from crud_views.lib.crispy import CrispyModelForm, Column4, CrispyModelViewMixin, CrispyDeleteForm
from crud_views.lib.table import Table, LinkDetailColumn
from crud_views.lib.views import (
    DetailViewPermissionRequired,
    UpdateViewPermissionRequired,
    CreateViewPermissionRequired,
    ListViewTableMixin,
    DeleteViewPermissionRequired,
    ListViewPermissionRequired,
    MessageMixin,
)
from crud_views.lib.viewset import ViewSet
from crud_views_workflow.lib.forms import WorkflowForm
from crud_views_workflow.lib.views import WorkflowViewPermissionRequired

cv_campaign = ViewSet(model=Campaign, name="campaign", icon_header="fa-solid fa-bullhorn")


class CampaignForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Campaign
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class CampaignUpdateForm(CampaignForm):
    submit_label = "Update"


class CampaignTable(Table):
    id = LinkDetailColumn()
    name = tables.Column(attrs=Table.ca.w70)
    state = tables.Column(accessor="state_badge", attrs=Table.ca.w20)


class CampaignListView(ListViewTableMixin, ListViewPermissionRequired):
    cv_viewset = cv_campaign
    cv_list_actions = ["detail", "workflow", "update", "delete"]
    table_class = CampaignTable


class CampaignDetailView(DetailViewPermissionRequired):
    cv_viewset = cv_campaign
    cv_context_actions = ["home", "detail", "update", "workflow", "delete"]

    cv_property_display = [
        {
            "title": "Properties",
            "icon": "megaphone",
            "description": "Campaign attributes",
            "properties": [
                "id",
                "name",
                PropertyConfig(path="state_badge", title=_("State"), detail="Campaign workflow state"),
            ],
        },
    ]


class CampaignUpdateView(CrispyModelViewMixin, MessageMixin, UpdateViewPermissionRequired):
    cv_context_actions = ["home", "detail", "update", "workflow", "delete"]
    form_class = CampaignUpdateForm
    cv_viewset = cv_campaign


class CampaignCreateView(CrispyModelViewMixin, MessageMixin, CreateViewPermissionRequired):
    form_class = CampaignForm
    cv_viewset = cv_campaign


class CampaignDeleteView(CrispyModelViewMixin, MessageMixin, DeleteViewPermissionRequired):
    cv_context_actions = ["home", "detail", "update", "workflow", "delete"]
    form_class = CrispyDeleteForm
    cv_viewset = cv_campaign


class CampaignWorkflowForm(WorkflowForm):
    class Meta(WorkflowForm.Meta):
        model = Campaign


class CampaignWorkflowView(CrispyModelViewMixin, MessageMixin, WorkflowViewPermissionRequired):
    cv_context_actions = ["list", "detail", "edit", "workflow"]
    cv_viewset = cv_campaign
    form_class = CampaignWorkflowForm



from django.db import models
from django.utils.translation import gettext as _
from django_fsm import FSMField, transition

from crud_views_workflow.lib.enums import WorkflowComment
from crud_views_workflow.lib.enums import BadgeEnum
from crud_views_workflow.lib.mixins import WorkflowModelMixin


class CampaignState(models.TextChoices):
    NEW = "new", _("New")
    ACTIVE = "active", _("Active")
    SUCCESS = "success", _("Success")
    CANCELED = "canceled", _("Cancelled")
    ERROR = "error", _("Error")


class Campaign(WorkflowModelMixin, models.Model):
    STATE_CHOICES = CampaignState
    STATE_BADGES = {
        CampaignState.NEW: BadgeEnum.LIGHT,
        CampaignState.ACTIVE: BadgeEnum.INFO,
        CampaignState.SUCCESS: BadgeEnum.PRIMARY,
        CampaignState.CANCELED: BadgeEnum.WARNING,
        CampaignState.ERROR: BadgeEnum.DANGER,
    }

    TRANSITION_COMMENT_OPTIONAL = 1
    TRANSITION_COMMENT_REQUIRED = 2

    name = models.CharField(max_length=128)
    state = FSMField(default=CampaignState.NEW, choices=CampaignState.choices)

    def __str__(self):
        return self.name

    @transition(
        field=state,
        source=CampaignState.NEW,
        target=CampaignState.ACTIVE,
        on_error=CampaignState.ERROR,
        custom={"label": _("Activate"), "comment": WorkflowComment.NONE},
    )
    def wf_activate(self, request=None, by=None, comment=None):
        pass

    @transition(
        field=state,
        source=CampaignState.ACTIVE,
        target=CampaignState.SUCCESS,
        on_error=CampaignState.ERROR,
        custom={"label": _("Done"), "comment": WorkflowComment.OPTIONAL},
    )
    def wf_done(self, request=None, by=None, comment=None):
        pass

    @transition(
        field=state,
        source=CampaignState.NEW,
        target=CampaignState.CANCELED,
        on_error=CampaignState.ERROR,
        custom={"label": _("Cancel"), "comment": WorkflowComment.REQUIRED},
    )
    def wf_cancel_new(self, request=None, by=None, comment=None):
        pass

    @transition(
        field=state,
        source=CampaignState.ACTIVE,
        target=CampaignState.CANCELED,
        on_error=CampaignState.ERROR,
        custom={"label": _("Cancel"), "comment": WorkflowComment.REQUIRED},
    )
    def wf_cancel_active(self, request=None, by=None, comment=None):
        pass



import pytest
from django.test.client import Client

from crud_views.lib.viewset import ViewSet
from tests.test1.app.models import Author


@pytest.mark.django_db
def test_ordered_views_registered(cv_author: ViewSet):
    """Up and down action views should be registered on the Author viewset."""
    assert cv_author.has_view("up"), "ViewSet missing view: up"
    assert cv_author.has_view("down"), "ViewSet missing view: down"


@pytest.mark.django_db
def test_ordered_url_patterns(cv_author: ViewSet):
    """URL patterns should include up and down routes."""
    patterns = cv_author.urlpatterns
    routes = [p.name for p in patterns]
    assert "author-up" in routes
    assert "author-down" in routes


@pytest.mark.django_db
def test_move_up(client_user_author_change: Client, cv_author):
    """Moving the second author up should swap positions."""
    a1 = Author.objects.create(first_name="First", last_name="Author")
    a2 = Author.objects.create(first_name="Second", last_name="Author")

    # a1 order=0, a2 order=1
    assert a1.order < a2.order

    # POST to move a2 up
    response = client_user_author_change.post(f"/author/{a2.pk}/up/")
    assert response.status_code == 302

    a1.refresh_from_db()
    a2.refresh_from_db()
    assert a2.order < a1.order


@pytest.mark.django_db
def test_move_up_first_item_noop(client_user_author_change: Client, cv_author):
    """Moving the first item up should be a no-op (already at top)."""
    a1 = Author.objects.create(first_name="First", last_name="Author")
    Author.objects.create(first_name="Second", last_name="Author")

    original_order = a1.order

    response = client_user_author_change.post(f"/author/{a1.pk}/up/")
    assert response.status_code == 302

    a1.refresh_from_db()
    assert a1.order == original_order


@pytest.mark.django_db
def test_move_down(client_user_author_change: Client, cv_author):
    """Moving the first author down should swap positions."""
    a1 = Author.objects.create(first_name="First", last_name="Author")
    a2 = Author.objects.create(first_name="Second", last_name="Author")

    assert a1.order < a2.order

    # POST to move a1 down
    response = client_user_author_change.post(f"/author/{a1.pk}/down/")
    assert response.status_code == 302

    a1.refresh_from_db()
    a2.refresh_from_db()
    assert a1.order > a2.order


@pytest.mark.django_db
def test_move_down_last_item_noop(client_user_author_change: Client, cv_author):
    """Moving the last item down should be a no-op (already at bottom)."""
    Author.objects.create(first_name="First", last_name="Author")
    a2 = Author.objects.create(first_name="Second", last_name="Author")

    original_order = a2.order

    response = client_user_author_change.post(f"/author/{a2.pk}/down/")
    assert response.status_code == 302

    a2.refresh_from_db()
    assert a2.order == original_order


@pytest.mark.django_db
def test_ordered_redirects_to_list(client_user_author_change: Client, cv_author):
    """After up/down, user should be redirected to the list view."""
    a1 = Author.objects.create(first_name="First", last_name="Author")

    response = client_user_author_change.post(f"/author/{a1.pk}/up/")
    assert response.status_code == 302
    assert response.url == "/author/"


@pytest.mark.django_db
def test_ordered_requires_permission(client: Client, cv_author, user_a):
    """User without change permission should get 403."""
    client.force_login(user_a)
    a1 = Author.objects.create(first_name="First", last_name="Author")

    response = client.post(f"/author/{a1.pk}/up/")
    assert response.status_code == 403

    response = client.post(f"/author/{a1.pk}/down/")
    assert response.status_code == 403


@pytest.mark.django_db
def test_ordered_anonymous_redirected(client: Client, cv_author):
    """Anonymous user should be redirected to login."""
    a1 = Author.objects.create(first_name="First", last_name="Author")

    response = client.post(f"/author/{a1.pk}/up/")
    assert response.status_code == 302
    assert "/accounts/login/" in response.url


@pytest.mark.django_db
def test_ordered_get_not_allowed(client_user_author_change: Client, cv_author):
    """GET should not be allowed on action views (POST only)."""
    a1 = Author.objects.create(first_name="First", last_name="Author")

    response = client_user_author_change.get(f"/author/{a1.pk}/up/")
    assert response.status_code == 405



from django.core.checks import Tags as DjangoTags
from django.core.checks import register

from crud_views.lib.settings import crud_views_settings
from crud_views.lib.viewset import ViewSet


class Tags(DjangoTags):
    """Do this if none of the existing tags work for you:
    https://docs.djangoproject.com/en/1.8/ref/checks/#builtin-tags
    """

    my_new_tag = "my_new_tag"


@register(Tags.my_new_tag)
def check_taggit_is_installed(app_configs=None, **kwargs):
    "Check that django-taggit is installed when usying myapp."

    errors = crud_views_settings.check_messages

    for check in ViewSet.checks_all():
        for message in check.messages():
            errors.append(message)
    return errors



"""
Tests for crud_views_workflow: WorkflowModelMixin, WorkflowView, WorkflowForm.
"""

import pytest
from django.test.client import Client

from tests.test1.app.models import Campaign, CampaignState
from crud_views_workflow.lib.enums import BadgeEnum
from crud_views_workflow.models import WorkflowInfo


# ---------------------------------------------------------------------------
# WorkflowModelMixin model-layer tests
# ---------------------------------------------------------------------------


@pytest.mark.django_db
def test_state_name_new(campaign_new):
    assert campaign_new.state_name == "New"


@pytest.mark.django_db
def test_state_name_active(campaign_active):
    assert campaign_active.state_name == "Active"


@pytest.mark.django_db
def test_state_name_success(campaign_success):
    assert campaign_success.state_name == "Success"


@pytest.mark.django_db
def test_get_state_name(campaign_new):
    assert campaign_new.get_state_name(CampaignState.NEW) == "New"
    assert campaign_new.get_state_name(CampaignState.ACTIVE) == "Active"
    assert campaign_new.get_state_name(CampaignState.SUCCESS) == "Success"
    assert campaign_new.get_state_name(CampaignState.CANCELED) == "Cancelled"


@pytest.mark.django_db
def test_state_badge_new(campaign_new):
    badge = campaign_new.state_badge
    assert "New" in badge
    assert "badge" in badge
    assert "light" in badge  # STATE_BADGES[NEW] = BadgeEnum.LIGHT


@pytest.mark.django_db
def test_get_state_badge_active(campaign_new):
    badge = campaign_new.get_state_badge(CampaignState.ACTIVE)
    assert "Active" in badge
    assert "info" in badge  # STATE_BADGES[ACTIVE] = BadgeEnum.INFO


@pytest.mark.django_db
def test_get_state_badge_unknown_uses_default(campaign_new):
    # State not in STATE_BADGES falls back to STATE_BADGE_DEFAULT badge class
    campaign_new.STATE_BADGES = {CampaignState.NEW: BadgeEnum.LIGHT}  # ERROR intentionally missing
    result = campaign_new.get_state_badge(CampaignState.ERROR)
    assert "Error" in result
    assert "badge" in result  # still renders a badge, not plain text
    assert "info" in result  # STATE_BADGE_DEFAULT = BadgeEnum.INFO


@pytest.mark.django_db
def test_workflow_possible_transitions_new(campaign_new, user_campaign_change):
    transitions = campaign_new.workflow_get_possible_transitions(user_campaign_change)
    names = [t[0] for t in transitions]
    assert "wf_activate" in names
    assert "wf_cancel_new" in names
    # wf_done is only available from ACTIVE
    assert "wf_done" not in names


@pytest.mark.django_db
def test_workflow_possible_transitions_active(campaign_active, user_campaign_change):
    transitions = campaign_active.workflow_get_possible_transitions(user_campaign_change)
    names = [t[0] for t in transitions]
    assert "wf_done" in names
    assert "wf_cancel_active" in names
    # wf_activate is only available from NEW
    assert "wf_activate" not in names


@pytest.mark.django_db
def test_workflow_possible_transitions_success_empty(campaign_success, user_campaign_change):
    transitions = campaign_success.workflow_get_possible_transitions(user_campaign_change)
    assert transitions == []


@pytest.mark.django_db
def test_workflow_has_transition_true(campaign_new, user_campaign_change):
    assert campaign_new.workflow_has_transition(user_campaign_change, "wf_activate") is True


@pytest.mark.django_db
def test_workflow_has_transition_false_wrong_state(campaign_new, user_campaign_change):
    # wf_done requires ACTIVE state
    assert campaign_new.workflow_has_transition(user_campaign_change, "wf_done") is False


@pytest.mark.django_db
def test_workflow_has_any_transition_true(campaign_new, user_campaign_change):
    assert campaign_new.workflow_has_any_possible_transition(user_campaign_change) is True


@pytest.mark.django_db
def test_workflow_has_any_transition_false_terminal(campaign_success, user_campaign_change):
    assert campaign_success.workflow_has_any_possible_transition(user_campaign_change) is False


@pytest.mark.django_db
def test_workflow_comment_kwargs_with_comment():
    result = Campaign.workflow_comment_kwargs("hello")
    assert result == {"comment": "hello"}


@pytest.mark.django_db
def test_workflow_comment_kwargs_without_comment():
    result = Campaign.workflow_comment_kwargs(None)
    assert result == {}


@pytest.mark.django_db
def test_workflow_comment_kwargs_empty_string():
    result = Campaign.workflow_comment_kwargs("")
    assert result == {}


@pytest.mark.django_db
def test_workflow_data_empty(campaign_new):
    assert campaign_new.workflow_data == []


@pytest.mark.django_db
def test_workflow_transition_label_map(campaign_new):
    label_map = campaign_new.workflow_get_transition_label_map
    assert "wf_activate" in label_map
    assert label_map["wf_activate"] == "Activate"
    assert "wf_done" in label_map
    assert label_map["wf_done"] == "Done"
    assert "wf_cancel_new" in label_map
    assert label_map["wf_cancel_new"] == "Cancel"


@pytest.mark.django_db
def test_workflow_get_transition_label(campaign_new):
    assert campaign_new.workflow_get_transition_label("wf_activate") == "Activate"
    assert campaign_new.workflow_get_transition_label("wf_done") == "Done"


@pytest.mark.django_db
def test_workflow_get_form_kwargs(campaign_new, user_campaign_change):
    kwargs = campaign_new.workflow_get_form_kwargs(user_campaign_change)
    assert "choices" in kwargs
    assert "transition_comments" in kwargs
    choice_names = [c[0] for c in kwargs["choices"]]
    assert "wf_activate" in choice_names
    assert "wf_cancel_new" in choice_names


@pytest.mark.django_db
def test_workflow_get_form_kwargs_labels(campaign_new, user_campaign_change):
    kwargs = campaign_new.workflow_get_form_kwargs(user_campaign_change)
    choices_dict = dict(kwargs["choices"])
    assert choices_dict["wf_activate"] == "Activate"
    assert choices_dict["wf_cancel_new"] == "Cancel"


@pytest.mark.django_db
def test_workflow_get_form_kwargs_comment_types(campaign_new, user_campaign_change):
    from crud_views_workflow.lib.enums import WorkflowComment

    kwargs = campaign_new.workflow_get_form_kwargs(user_campaign_change)
    tc = kwargs["transition_comments"]
    assert tc["wf_activate"] == WorkflowComment.NONE
    assert tc["wf_cancel_new"] == WorkflowComment.REQUIRED


def test_comment_default_is_none():
    """COMMENT_DEFAULT defaults to WorkflowComment.NONE."""
    from crud_views_workflow.lib.enums import WorkflowComment
    from crud_views_workflow.lib.mixins import WorkflowModelMixin

    assert WorkflowModelMixin.COMMENT_DEFAULT == WorkflowComment.NONE


@pytest.mark.django_db
def test_comment_default_fallback(campaign_new, user_campaign_change):
    """COMMENT_DEFAULT is used when a transition omits 'comment' from custom."""
    from unittest.mock import MagicMock, patch
    from crud_views_workflow.lib.enums import WorkflowComment

    mock_transition = MagicMock()
    mock_transition.name = "wf_test"
    mock_transition.custom = {"label": "Test"}  # no 'comment' key

    with patch.object(campaign_new, "get_available_user_state_transitions", return_value=[mock_transition]):
        transitions = campaign_new.workflow_get_possible_transitions(user_campaign_change)
        assert transitions[0][2] == WorkflowComment.NONE  # default


@pytest.mark.django_db
def test_comment_default_override(campaign_new, user_campaign_change):
    """Overriding COMMENT_DEFAULT changes the fallback for transitions without an explicit comment."""
    from unittest.mock import MagicMock, patch
    from crud_views_workflow.lib.enums import WorkflowComment

    mock_transition = MagicMock()
    mock_transition.name = "wf_test"
    mock_transition.custom = {"label": "Test"}  # no 'comment' key

    campaign_new.COMMENT_DEFAULT = WorkflowComment.OPTIONAL
    with patch.object(campaign_new, "get_available_user_state_transitions", return_value=[mock_transition]):
        transitions = campaign_new.workflow_get_possible_transitions(user_campaign_change)
        assert transitions[0][2] == WorkflowComment.OPTIONAL


# ---------------------------------------------------------------------------
# WorkflowView HTTP tests
# ---------------------------------------------------------------------------


@pytest.mark.django_db
def test_workflow_view_get(client_user_campaign_change: Client, campaign_new):
    response = client_user_campaign_change.get(f"/campaign/{campaign_new.pk}/workflow/")
    assert response.status_code == 200


@pytest.mark.django_db
def test_workflow_view_get_shows_transitions(client_user_campaign_change: Client, campaign_new):
    response = client_user_campaign_change.get(f"/campaign/{campaign_new.pk}/workflow/")
    assert response.status_code == 200
    assert response.context["has_workflow_choices"] is True


@pytest.mark.django_db
def test_workflow_view_get_terminal_state_no_choices(client_user_campaign_change: Client, campaign_success):
    response = client_user_campaign_change.get(f"/campaign/{campaign_success.pk}/workflow/")
    assert response.status_code == 200
    assert response.context["has_workflow_choices"] is False


@pytest.mark.django_db
def test_workflow_view_post_activate(client_user_campaign_change: Client, campaign_new):
    pk = campaign_new.pk
    response = client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_activate",
            "comment": "",
        },
    )
    assert response.status_code == 302
    campaign_new.refresh_from_db()
    assert campaign_new.state == CampaignState.ACTIVE


@pytest.mark.django_db
def test_workflow_view_post_creates_workflow_info(client_user_campaign_change: Client, campaign_new):
    pk = campaign_new.pk
    client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_activate",
            "comment": "",
        },
    )
    assert WorkflowInfo.objects.filter(
        workflow_object_pk=str(pk),
        transition="wf_activate",
        state_old=CampaignState.NEW,
        state_new=CampaignState.ACTIVE,
    ).exists()


@pytest.mark.django_db
def test_workflow_view_post_required_comment_missing(client_user_campaign_change: Client, campaign_new):
    """POST with required-comment transition but empty comment → form error, state unchanged."""
    pk = campaign_new.pk
    response = client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_cancel_new",
            "comment": "",
        },
    )
    assert response.status_code == 200  # form invalid, re-render
    campaign_new.refresh_from_db()
    assert campaign_new.state == CampaignState.NEW  # state unchanged


@pytest.mark.django_db
def test_workflow_view_post_required_comment_provided(client_user_campaign_change: Client, campaign_new):
    pk = campaign_new.pk
    response = client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_cancel_new",
            "comment": "Need to cancel this campaign",
        },
    )
    assert response.status_code == 302
    campaign_new.refresh_from_db()
    assert campaign_new.state == CampaignState.CANCELED
    info = WorkflowInfo.objects.get(workflow_object_pk=str(pk), transition="wf_cancel_new")
    assert info.comment == "Need to cancel this campaign"


@pytest.mark.django_db
def test_workflow_view_post_optional_comment_without(client_user_campaign_change: Client, campaign_active):
    pk = campaign_active.pk
    response = client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_done",
            "comment": "",
        },
    )
    assert response.status_code == 302
    campaign_active.refresh_from_db()
    assert campaign_active.state == CampaignState.SUCCESS


@pytest.mark.django_db
def test_workflow_view_post_optional_comment_with(client_user_campaign_change: Client, campaign_active):
    pk = campaign_active.pk
    response = client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_done",
            "comment": "Campaign completed successfully",
        },
    )
    assert response.status_code == 302
    campaign_active.refresh_from_db()
    assert campaign_active.state == CampaignState.SUCCESS
    info = WorkflowInfo.objects.get(workflow_object_pk=str(pk), transition="wf_done")
    assert info.comment == "Campaign completed successfully"


@pytest.mark.django_db
def test_workflow_view_post_unavailable_transition_form_error(client_user_campaign_change: Client, campaign_new):
    """Submitting a transition not available for the current state is rejected by form validation."""
    pk = campaign_new.pk
    response = client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_done",  # only available from ACTIVE, not NEW
            "comment": "",
        },
    )
    # form validation rejects the choice; view re-renders the form
    assert response.status_code == 200
    campaign_new.refresh_from_db()
    assert campaign_new.state == CampaignState.NEW  # state unchanged


@pytest.mark.django_db
def test_workflow_history_after_multiple_transitions(client_user_campaign_change: Client, campaign_new):
    pk = campaign_new.pk

    # Transition 1: Activate
    client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_activate",
            "comment": "",
        },
    )

    # Transition 2: Done (with comment)
    client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_done",
            "comment": "All done",
        },
    )

    campaign_new.refresh_from_db()
    assert campaign_new.state == CampaignState.SUCCESS

    assert WorkflowInfo.objects.filter(workflow_object_pk=str(pk)).count() == 2

    history = campaign_new.workflow_data
    assert len(history) == 2
    assert history[0]["transition"] == "wf_activate"
    assert history[0]["state_old"] == CampaignState.NEW
    assert history[0]["state_new"] == CampaignState.ACTIVE
    assert history[1]["transition"] == "wf_done"
    assert history[1]["state_old"] == CampaignState.ACTIVE
    assert history[1]["state_new"] == CampaignState.SUCCESS
    assert history[1]["comment"] == "All done"


@pytest.mark.django_db
def test_workflow_history_transition_labels(client_user_campaign_change: Client, campaign_new):
    pk = campaign_new.pk
    client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_activate",
            "comment": "",
        },
    )
    campaign_new.refresh_from_db()
    history = campaign_new.workflow_data
    assert history[0]["transition_label"] == "Activate"


@pytest.mark.django_db
def test_workflow_history_records_user(client_user_campaign_change: Client, campaign_new, user_campaign_change):
    pk = campaign_new.pk
    client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_activate",
            "comment": "",
        },
    )
    info = WorkflowInfo.objects.get(workflow_object_pk=str(pk), transition="wf_activate")
    assert info.user == user_campaign_change


@pytest.mark.django_db
def test_workflow_info_null_comment_when_empty(client_user_campaign_change: Client, campaign_new):
    """A blank comment is stored as NULL in WorkflowInfo."""
    pk = campaign_new.pk
    client_user_campaign_change.post(
        f"/campaign/{pk}/workflow/",
        {
            "transition": "wf_activate",
            "comment": "",
        },
    )
    info = WorkflowInfo.objects.get(workflow_object_pk=str(pk), transition="wf_activate")
    assert info.comment is None


@pytest.mark.django_db
def test_workflow_view_get_contains_campaign_name(client_user_campaign_change: Client, campaign_new):
    response = client_user_campaign_change.get(f"/campaign/{campaign_new.pk}/workflow/")
    assert response.status_code == 200
    assert campaign_new.name.encode() in response.content


# ---------------------------------------------------------------------------
# WorkflowViewPermissionRequired: access control tests
# ---------------------------------------------------------------------------


@pytest.mark.django_db
def test_workflow_permission_required_allows_change_user(client_user_campaign_change: Client, campaign_new):
    """User with change permission can access the workflow view."""
    response = client_user_campaign_change.get(f"/campaign/{campaign_new.pk}/workflow/")
    assert response.status_code == 200


@pytest.mark.django_db
def test_workflow_permission_required_denies_view_only_user(client_user_campaign_view: Client, campaign_new):
    """User with only view permission is denied access (403)."""
    response = client_user_campaign_view.get(f"/campaign/{campaign_new.pk}/workflow/")
    assert response.status_code == 403


@pytest.mark.django_db
def test_workflow_permission_required_denies_post_view_only_user(client_user_campaign_view: Client, campaign_new):
    """User with only view permission cannot POST a transition (403)."""
    response = client_user_campaign_view.post(
        f"/campaign/{campaign_new.pk}/workflow/",
        {
            "transition": "wf_activate",
            "comment": "",
        },
    )
    assert response.status_code == 403
    campaign_new.refresh_from_db()
    assert campaign_new.state == CampaignState.NEW  # state unchanged


@pytest.mark.django_db
def test_workflow_permission_required_redirects_anonymous(client: Client, campaign_new):
    """Anonymous user is redirected to the login page."""
    response = client.get(f"/campaign/{campaign_new.pk}/workflow/")
    assert response.status_code == 302
    assert "/login" in response["Location"]


# ---------------------------------------------------------------------------
# WorkflowView.checks() system checks
# ---------------------------------------------------------------------------


def _errors(view_cls):
    """Collect all error messages from a view's checks()."""
    return [msg for chk in view_cls.checks() for msg in chk.messages()]


def _error_ids(view_cls):
    return {e.id for e in _errors(view_cls)}


def test_workflow_view_checks_pass():
    """A correctly configured WorkflowView produces no check errors."""
    from tests.test1.app.views import CampaignWorkflowView

    assert _errors(CampaignWorkflowView) == []


def test_workflow_view_checks_missing_form_class():
    """Missing form_class yields E230."""
    from crud_views_workflow.lib.views import WorkflowView

    class NoFormView(WorkflowView):
        cv_key = "wf"
        cv_path = "wf"
        # form_class deliberately left as None (the WorkflowView default)

    assert "viewset.E230" in _error_ids(NoFormView)


def test_workflow_view_checks_transition_label_none():
    """cv_transition_label=None yields E231."""
    from crud_views_workflow.lib.views import WorkflowView
    from crud_views_workflow.lib.forms import WorkflowForm

    class NoLabelView(WorkflowView):
        cv_key = "wf"
        cv_path = "wf"
        form_class = WorkflowForm
        cv_transition_label = None

    assert "viewset.E231" in _error_ids(NoLabelView)


def test_workflow_view_checks_comment_label_none():
    """cv_comment_label=None yields E232."""
    from crud_views_workflow.lib.views import WorkflowView
    from crud_views_workflow.lib.forms import WorkflowForm

    class NoCommentLabelView(WorkflowView):
        cv_key = "wf"
        cv_path = "wf"
        form_class = WorkflowForm
        cv_comment_label = None

    assert "viewset.E232" in _error_ids(NoCommentLabelView)


def test_workflow_view_checks_model_not_mixin():
    """Model that does not extend WorkflowModelMixin yields E233."""
    from unittest.mock import MagicMock
    from crud_views_workflow.lib.views import WorkflowView
    from crud_views_workflow.lib.forms import WorkflowForm

    class PlainModel:
        pass

    class BadModelView(WorkflowView):
        cv_key = "wf"
        cv_path = "wf"
        form_class = WorkflowForm

    mock_viewset = MagicMock()
    mock_viewset.model = PlainModel
    BadModelView.cv_viewset = mock_viewset

    assert "viewset.E233" in _error_ids(BadModelView)


def test_workflow_view_checks_model_missing_state_choices():
    """WorkflowModelMixin model without STATE_CHOICES set yields E234."""
    from unittest.mock import MagicMock
    from crud_views_workflow.lib.views import WorkflowView
    from crud_views_workflow.lib.forms import WorkflowForm
    from crud_views_workflow.lib.mixins import WorkflowModelMixin

    class NoChoicesModel(WorkflowModelMixin):
        STATE_CHOICES = None
        STATE_BADGES = {"active": BadgeEnum.INFO}

    class NoChoicesView(WorkflowView):
        cv_key = "wf"
        cv_path = "wf"
        form_class = WorkflowForm

    mock_viewset = MagicMock()
    mock_viewset.model = NoChoicesModel
    NoChoicesView.cv_viewset = mock_viewset

    assert "viewset.E234" in _error_ids(NoChoicesView)


def test_workflow_view_checks_model_missing_state_badges():
    """WorkflowModelMixin model without STATE_BADGES set yields E235."""
    from unittest.mock import MagicMock
    from crud_views_workflow.lib.views import WorkflowView
    from crud_views_workflow.lib.forms import WorkflowForm
    from crud_views_workflow.lib.mixins import WorkflowModelMixin

    class NoBadgesModel(WorkflowModelMixin):
        STATE_CHOICES = object()
        STATE_BADGES = None

    class NoBadgesView(WorkflowView):
        cv_key = "wf"
        cv_path = "wf"
        form_class = WorkflowForm

    mock_viewset = MagicMock()
    mock_viewset.model = NoBadgesModel
    NoBadgesView.cv_viewset = mock_viewset

    assert "viewset.E235" in _error_ids(NoBadgesView)



from typing import List

from crispy_forms.layout import Row
from django.forms import RadioSelect, ChoiceField, CharField
from django.forms.widgets import Textarea
from django.utils.translation import gettext_lazy as _

from crud_views.lib.crispy import CrispyForm, Column8
from .enums import WorkflowComment


class WorkflowForm(CrispyForm):
    """
    Workflow form
    """

    class Meta:
        fields = ["transition", "comment"]

    submit_label: str = _("Process Workflow Step")

    transition = ChoiceField(
        choices=tuple(),
        widget=RadioSelect,
        label="LABEL-PLACEHOLDER",
        help_text="HELP-PLACEHOLDER",
        required=True,
    )
    comment = CharField(
        label="LABEL-PLACEHOLDER",
        help_text="HELP-PLACEHOLDER",
        required=False,  # will be set dynamically
        widget=Textarea(
            attrs={
                "rows": 4,
                "placeholder": _("Add your comment…"),
            }
        ),
    )

    def get_layout_fields(self):
        return Row(Column8("transition")), Row(Column8("comment"))

    def __init__(self, *args, choices: List[tuple[str, str]], transition_comments: dict, **kwargs):
        super().__init__(*args, **kwargs)
        self.transition_comments = transition_comments
        self.transition_comment = (
            max(transition_comments.values()) if transition_comments.values() else WorkflowComment.NONE
        )

        # hide the comment field if not required
        if self.transition_comment == WorkflowComment.NONE:
            field_comment = self.fields["comment"]
            field_comment.widget = field_comment.hidden_widget()

        transition = self.fields["transition"]
        transition.choices = choices
        transition.widget.use_fieldset = False

    def clean(self):
        cleaned_data = super().clean()
        transition = cleaned_data.get("transition")
        comment = cleaned_data.get("comment")
        if isinstance(comment, str):
            comment = comment.strip()
        transition_comment = self.transition_comments.get(transition)
        if transition_comment == WorkflowComment.REQUIRED and not comment:
            self.add_error("comment", _("Comment is required for this transition"))



import pytest
from django.test.client import Client

from tests.test1.app.models import Author


@pytest.mark.django_db
def test_create_author(client_user_author_add: Client, cv_author):
    # GET create form
    response = client_user_author_add.get("/author/create/")
    assert response.status_code == 200

    # POST valid data
    response = client_user_author_add.post(
        "/author/create/",
        {
            "first_name": "Isaac",
            "last_name": "Asimov",
            "pseudonym": "",
        },
    )
    assert response.status_code == 302
    assert response.url == "/author/"

    # Verify author created in DB
    assert Author.objects.filter(first_name="Isaac", last_name="Asimov").exists()


@pytest.mark.django_db
def test_update_author(client_user_author_change: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    # GET update form
    response = client_user_author_change.get(f"/author/{pk}/update/")
    assert response.status_code == 200

    # POST updated data
    response = client_user_author_change.post(
        f"/author/{pk}/update/",
        {
            "first_name": "Douglas",
            "last_name": "Adams",
            "pseudonym": "DNA",
        },
    )
    assert response.status_code == 302
    assert response.url == "/author/"

    # Verify author updated in DB
    author_douglas_adams.refresh_from_db()
    assert author_douglas_adams.pseudonym == "DNA"


@pytest.mark.django_db
def test_delete_author(client_user_author_delete: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    # GET delete confirmation
    response = client_user_author_delete.get(f"/author/{pk}/delete/")
    assert response.status_code == 200

    # POST to confirm deletion (form requires confirm checkbox)
    response = client_user_author_delete.post(f"/author/{pk}/delete/", {"confirm": True})
    assert response.status_code == 302
    assert response.url == "/author/"

    # Verify author deleted from DB
    assert not Author.objects.filter(pk=pk).exists()


@pytest.mark.django_db
def test_detail_author(client_user_author_view: Client, cv_author, author_douglas_adams):
    pk = author_douglas_adams.pk

    # GET detail page
    response = client_user_author_view.get(f"/author/{pk}/detail/")
    assert response.status_code == 200

    # Verify content contains author name
    content = response.content.decode("utf-8")
    assert "Douglas" in content
    assert "Adams" in content



from crispy_forms.layout import Row
from crud_views.lib.crispy import Column4, Column8
from polymorphic.formsets import (
    polymorphic_inlineformset_factory,
    PolymorphicFormSetChild,
    BasePolymorphicInlineFormSet,
)

from app.models.poly import Poly, PolyAnswer, PolyAnswerText, PolyAnswerNumber
from crud_views.lib.formsets.inline_formset import CrispyInlineModelForm


class PolyAnswerTextForm(CrispyInlineModelForm):
    class Meta:
        model = PolyAnswerText
        fields = ["answer", "text"]

    def get_layout_fields(self):
        return Row(Column4("answer"), Column8("text"))  # , "id"


PolyAnswerTextFormSetChild = PolymorphicFormSetChild(
    PolyAnswerText,
    form=PolyAnswerTextForm,
    fields=None,
    exclude=None,
    formfield_callback=None,
    widgets=None,
    localized_fields=None,
    labels=None,
    help_texts=None,
    error_messages=None,
)


class PolyAnswerNumberForm(CrispyInlineModelForm):
    class Meta:
        model = PolyAnswerNumber
        fields = [
            # "id",
            "answer",
            "number",
        ]

    def get_layout_fields(self):
        return Row(Column8("answer"), Column4("number"))  # , "id"


PolyAnswerNumberFormSetChild = PolymorphicFormSetChild(
    PolyAnswerNumber,
    form=PolyAnswerNumberForm,
    fields=None,
    exclude=None,
    formfield_callback=None,
    widgets=None,
    localized_fields=None,
    labels=None,
    help_texts=None,
    error_messages=None,
)


# todo: move this to module
class PolymorphicInlineFormSet(BasePolymorphicInlineFormSet):
    def __init__(self, formset, parent_form, *args, **kwargs):
        self.formset = formset
        self.parent_form = parent_form
        super().__init__(*args, **kwargs)


PolyAnswerFormSet = polymorphic_inlineformset_factory(
    parent_model=Poly,
    model=PolyAnswer,
    formset=PolymorphicInlineFormSet,
    formset_children=[PolyAnswerTextFormSetChild, PolyAnswerNumberFormSetChild],
    fields="__all__",
    extra=0,  # polymorphic inlines cannot be created currently
    can_delete=False,
    can_order=False,
)



from collections import OrderedDict

from django.contrib.auth.mixins import PermissionRequiredMixin
from django.views import generic

from crud_views.lib.settings import crud_views_settings
from crud_views.lib.view import CrudView


class ManageView(PermissionRequiredMixin, CrudView, generic.TemplateView):
    template_name = "crud_views/view_manage.html"

    cv_pk: bool = False  # does not need primary key
    cv_key = "manage"
    cv_path = "manage"
    cv_object = False

    cv_context_actions = crud_views_settings.manage_context_actions

    # texts and labels
    cv_header_template: str | None = "crud_views/snippets/header/manage.html"
    cv_paragraph_template: str | None = "crud_views/snippets/paragraph/manage.html"
    cv_action_label_template: str | None = "crud_views/snippets/action/manage.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/manage.html"

    # icons
    cv_icon_action = "fa-solid fa-gear"
    cv_icon_header = "fa-solid fa-gear"

    def has_permission(self):
        """
        Currently manage views are only attached to ViewSets via a global switch in settings
        """
        return True

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        permissions = self.cv_viewset.permissions
        rows = []
        for short, long in permissions.items():
            rows.append(dict(viewset=short, django=long, has_permission=self.request.user.has_perm(long)))
        views = self.get_view_data()
        context.update({"cv": self.cv_viewset, "data": rows, "views": views})
        return context

    def get_view_data(self):
        data = OrderedDict()
        for key, view in self.cv_viewset.get_all_views().items():
            view_data = OrderedDict(
                {
                    "base": OrderedDict(
                        {
                            "class": str(view.__class__),
                            "cv_key": view.cv_key,
                            "cv_path": view.cv_path,
                            "cv_backend_only": view.cv_backend_only,
                            "cv_list_actions": view.cv_list_actions,
                            "cv_list_action_method": view.cv_list_action_method,
                            "cv_context_actions": view.cv_context_actions,
                            "cv_home_key": view.cv_home_key,
                            "cv_success_key": view.cv_success_key,
                            "cv_cancel_key": view.cv_cancel_key,
                            "cv_parent_key": view.cv_parent_key,
                        }
                    ),
                    "templates": OrderedDict(
                        {
                            "cv_header_template": view.cv_header_template,
                            "cv_header_template_code": view.cv_header_template_code,
                            "cv_paragraph_template": view.cv_paragraph_template,
                            "cv_paragraph_template_code": view.cv_paragraph_template_code,
                            "cv_action_label_template": view.cv_action_label_template,
                            "cv_action_label_template_code": view.cv_action_label_template_code,
                            "cv_action_short_label_template": view.cv_action_short_label_template,
                            "cv_action_short_label_template_code": view.cv_action_short_label_template_code,
                        }
                    ),
                    "icons": OrderedDict(
                        {
                            "cv_icon_action": view.cv_icon_action,
                            "cv_icon_header": view.cv_icon_header,
                        }
                    ),
                }
            )
            data[key] = view_data
        return data



from django.views.generic.base import TemplateResponseMixin, View

from crud_views.lib.views.mixins import CrudViewProcessFormMixin
from django.views.generic.detail import DetailView
from django.views.generic.edit import FormMixin

from crud_views.lib.settings import crud_views_settings
from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin


class CustomFormView(CrudViewProcessFormMixin, CrudView, FormMixin, DetailView):
    """
    Object based view with custom form.
    Note: you have to set:
        - cv_key
        - cv_path
    """

    template_name = "crud_views/view_custom_form.html"
    cv_context_actions = crud_views_settings.detail_context_actions

    def cv_form_valid(self, context: dict):
        """
        Handle valid form
        """
        pass

    def cv_form_valid_hook(self, context: dict):
        """
        Handle valid form hook
        """
        pass


class CustomFormViewPermissionRequired(CrudViewPermissionRequiredMixin, CustomFormView):  # this file
    cv_permission = "view"


class CustomFormNoObjectView(CrudViewProcessFormMixin, CrudView, FormMixin, TemplateResponseMixin, View):
    """
    Non object based view with custom form.
    Note: you have to set:
        - cv_key
        - cv_path
    """

    template_name = "crud_views/view_custom_form.html"
    cv_context_actions = crud_views_settings.detail_context_actions

    def get(self, request, *args, **kwargs):
        context = self.get_context_data()
        return self.render_to_response(context)

    def cv_form_valid(self, context: dict):
        """
        Handle valid form
        """
        pass

    def cv_form_valid_hook(self, context: dict):
        """
        Handle valid form hook
        """
        pass


class CustomFormNoObjectViewPermissionRequired(CrudViewPermissionRequiredMixin, CustomFormNoObjectView):  # this file
    cv_permission = "view"



from typing import Type

import django_tables2 as tables

from .attrs import ColAttr
from .columns import ActionColumn


class TableWithViewContext(tables.Table):
    """
    Table with view context
    """

    ca: Type[ColAttr] = ColAttr

    def __init__(self, *args, **kwargs):
        if "view" not in kwargs:
            raise ValueError(f"view not set in {self.__class__}")
        view = kwargs.pop("view")
        super().__init__(*args, **kwargs)
        self.view = view


class Table(TableWithViewContext):
    """
    Table with action column
    """

    actions = ActionColumn()

    def __init__(self, *args, **kwargs):
        if "sequence" not in kwargs:
            # as default make sure actions are at the end
            kwargs["sequence"] = ("...", "actions")
        super().__init__(*args, **kwargs)

    @staticmethod
    def order_actions(queryset, is_descending):
        """
        do not order actions
        """
        return queryset, True



from crud_views.lib.views.mixins import CrudViewProcessFormMixin
from django.views import generic

from crud_views.lib.settings import crud_views_settings
from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin


class DeleteView(CrudViewProcessFormMixin, CrudView, generic.DeleteView):
    template_name = "crud_views/view_delete.html"

    cv_key = "delete"
    cv_path = "delete"
    cv_success_key = "list"
    cv_context_actions = crud_views_settings.delete_context_actions

    # texts and labels
    cv_header_template: str | None = "crud_views/snippets/header/delete.html"
    cv_paragraph_template: str | None = "crud_views/snippets/paragraph/delete.html"
    cv_action_label_template: str | None = "crud_views/snippets/action/delete.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/delete.html"

    # icons
    cv_icon_action = "fa-regular fa-trash-can"

    # messages
    cv_message_template: str | None = "crud_views/snippets/message/delete.html"

    def cv_form_valid(self, context: dict):
        """
        Handle valid form, delete the object
        """
        self.object.delete()


class DeleteViewPermissionRequired(CrudViewPermissionRequiredMixin, DeleteView):  # this file
    cv_permission = "delete"



from tests.test1.app.views import cv_author, cv_publisher, cv_book, cv_vehicle, cv_campaign

urlpatterns = []
urlpatterns += cv_author.urlpatterns
urlpatterns += cv_publisher.urlpatterns
urlpatterns += cv_book.urlpatterns
urlpatterns += cv_vehicle.urlpatterns
urlpatterns += cv_campaign.urlpatterns



from typing import Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from crud_views.lib.view.base import CrudView

from django.db.models import Model
from pydantic import BaseModel, field_validator


class ViewContext(BaseModel, arbitrary_types_allowed=True):
    """
    A container that is passed to methods of CrudView with the context:
        - views
        - object
    """

    view: "CrudView"
    object: Model | None = None

    @field_validator("object", mode="before")
    @classmethod
    def validate_object(cls, value: Any) -> Any:
        """
        When passed from template, an empty string is passed as object
        """
        if value == "":
            return None
        else:
            return value

    def to_dict(self) -> Dict[str, Any]:
        return {"object": self.object}

    @property
    def router_name(self) -> str:
        return self.view.request.resolver_match.url_name



from django.core.exceptions import ValidationError
from django.db import models
from ordered_model.models import OrderedModel


def validate_questionmark(value):
    if "?" in value:
        raise ValidationError(
            "%(value)s must not contain a questionmark",
            params={"value": value},
        )


def validate_alpha(value):
    if not value.isalpha():
        raise ValidationError(
            "%(value)s is not an alpha",
            params={"value": value},
        )


class Question(models.Model):
    question = models.CharField(max_length=100, validators=[validate_questionmark])

    def __str__(self):
        return f"{self.question}"


class QuestionTag(models.Model):
    question = models.ForeignKey(Question, on_delete=models.CASCADE)
    tag = models.CharField(max_length=100)


class QuestionChoice(OrderedModel):
    question = models.ForeignKey(Question, on_delete=models.CASCADE)
    choice = models.CharField(max_length=100, validators=[validate_alpha])
    help_text = models.CharField(max_length=100, blank=True, null=True)

    def __str__(self):
        return f"{self.choice}"


class QuestionChoiceTag(OrderedModel):
    choice = models.ForeignKey(QuestionChoice, on_delete=models.CASCADE)
    tag = models.CharField(max_length=100)


class QuestionChoiceTagAnnotation(OrderedModel):
    tag = models.ForeignKey(QuestionChoiceTag, on_delete=models.CASCADE)
    annotation = models.CharField(max_length=100)



# Generated by Django 4.2 on 2026-02-20 10:26

from django.db import migrations
import django_fsm


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0004_campaign"),
    ]

    operations = [
        migrations.AlterField(
            model_name="campaign",
            name="state",
            field=django_fsm.FSMField(
                choices=[
                    ("new", "New"),
                    ("active", "Created"),
                    ("success", "Error"),
                    ("canceled", "Assigned"),
                    ("error", "Error"),
                ],
                default="new",
                max_length=50,
            ),
        ),
    ]



from typing import List

from crispy_forms.bootstrap import FormActions
from crispy_forms.helper import FormHelper
from crispy_forms.layout import Button, Layout, Row, Submit, LayoutObject, BaseInput
from django.forms import ModelForm, Form, BooleanField
from django.utils.translation import gettext_lazy as _

from crud_views.lib.crispy import Column4


class CrispyFormMixin:
    """
    CrispyForm helps to application DRY:

        class FooForm(CrispyModelForm):
            class Meta:
                model = Foo
                fields = ["foo", "bar"]

            def get_layout_fields(self) -> LayoutObject | BaseInput | List[LayoutObject | BaseInput]:
                return [
                    Row(
                        Column6("foo"), Column6("bar"),
                    )
                ]

    The for has FormActions element with two buttons:
        - save
        - cancel, which is linked to the cancel key defined the CrudView

    The mixin for CrudView(s) adds the view context cv_view to the form. CrispyModelViewMixin,
    when added to a CrudView, sets this extra argument for the form in get_form_kwargs.
    """

    submit_label: str = _("Save")

    def __init__(self, cv_view, *args, **kwargs):
        self.cv_view = cv_view
        super().__init__(*args, **kwargs)

    @property
    def helper(self) -> FormHelper:
        helper = FormHelper()
        fields = self.get_layout_fields()
        if not isinstance(fields, (list, tuple)):
            fields = [fields]
        helper.layout = Layout(*fields, self.get_form_actions())
        return helper

    def get_layout_fields(self) -> LayoutObject | BaseInput | List[LayoutObject | BaseInput]:
        raise NotImplementedError

    def get_form_actions(self) -> FormActions:
        return FormActions(self.get_submit_button(), self.get_cancel_button())

    def get_submit_button_kwargs(self) -> dict:
        return {"name": "submit", "value": self.submit_label}

    def get_submit_button(self) -> Submit:
        """
        Submit button for FormActions
        """
        return Submit(**self.get_submit_button_kwargs())

    def get_cancel_button_kwargs(self) -> dict:
        request = self.cv_view.request
        # get the object from the form instance or the view context object
        obj = getattr(self, "instance", getattr(self.cv_view, "object", None))
        context = self.cv_view.get_cancel_button_context(obj=obj, user=request.user, request=request)
        url = context["cv_url"]
        onclick = f"location.href='{url}';return false;"
        return {
            "name": "reset",
            "value": context["cv_action_label"],
            "onclick": onclick,
            "css_class": "btn btn-secondary",
        }

    def get_cancel_button(self) -> Button:
        return Button(**self.get_cancel_button_kwargs())


class CrispyModelForm(CrispyFormMixin, ModelForm):
    """
    Base class for ModelForm with crispy forms.
    This form takes the current viewset view as an argument and uses it to generate the cancel button.
    """

    pass


class CrispyForm(CrispyFormMixin, Form):
    """
    Base class for Form with crispy forms.
    This form takes the current viewset view as an argument and uses it to generate the cancel button.
    """

    pass


class CrispyDeleteForm(CrispyForm):
    """
    Default delete from with confirmation
    """

    confirm = BooleanField(label=_("Confirm deletion"))

    @property
    def helper(self):
        helper = FormHelper()
        helper.layout = Layout(
            Row(Column4("confirm")),
            FormActions(
                Submit("delete", "Delete"),
                self.get_cancel_button(),
            ),
        )
        return helper


class CrispyViewMixin:
    """
    This mixin makes sure that CrispyModelForm gets the cv_view argument
    """

    def get_form_kwargs(self):
        kwargs = super().get_form_kwargs()  # noqa
        form_class = self.get_form_class()  # noqa
        if issubclass(form_class, (CrispyModelForm, CrispyForm)):
            kwargs["cv_view"] = self
        return kwargs


# todo: remove this
class CrispyModelViewMixin(CrispyViewMixin):
    pass



from functools import cached_property
from typing import Any, List

from box import Box
from django.conf import settings
from django.core.checks import CheckMessage, Error
from django.template import TemplateDoesNotExist
from django.template.loader import get_template
from pydantic import BaseModel, PrivateAttr


class Default:
    pass


_messages = []


def default_list(*args, **kwargs) -> Any:
    return list()


def from_settings(name, default=None) -> Any:
    return getattr(settings, name, default)


class CrudViewsSettings(BaseModel):
    # basic
    extends: str = from_settings(
        "CRUD_VIEWS_EXTENDS",
    )
    manage_views_enabled: str = from_settings(
        "CRUD_VIEWS_MANAGE_VIEWS_ENABLED", default="debug_only"
    )  # no, yes, debug_only

    # session
    session_data_key: str = from_settings("CRUD_VIEWS_SESSION_DATA_KEY", "viewset")

    # filter
    filter_persistence: bool = from_settings("CRUD_VIEWS_FILTER_PERSISTENCE", default=True)
    filter_icon: str = from_settings("CRUD_VIEWS_FILTER_ICON", default="fa-solid fa-filter")
    filter_reset_button_css_class: str = from_settings(
        "CRUD_VIEWS_FILTER_RESET_BUTTON_CSS_CLASS", default="btn btn-secondary"
    )

    # view defaults
    list_actions: List[str] = from_settings("CRUD_VIEWS_LIST_ACTIONS", default=["detail", "update", "delete"])
    list_context_actions: List[str] = from_settings(
        "CRUD_VIEWS_LIST_CONTEXT_ACTIONS", default=["parent", "list", "filter", "create"]
    )
    detail_context_actions: List[str] = from_settings(
        "CRUD_VIEWS_DETAIL_CONTEXT_ACTIONS", default=["home", "detail", "update", "delete"]
    )
    create_context_actions: List[str] = from_settings("CRUD_VIEWS_CREATE_CONTEXT_ACTIONS", default=["home", "create"])
    update_context_actions: List[str] = from_settings(
        "CRUD_VIEWS_UPDATE_CONTEXT_ACTIONS", default=["home", "detail", "update", "delete"]
    )
    delete_context_actions: List[str] = from_settings(
        "CRUD_VIEWS_DELETE_CONTEXT_ACTIONS", default=["home", "detail", "update", "delete"]
    )
    manage_context_actions: List[str] = from_settings("CRUD_VIEWS_MANAGE_CONTEXT_ACTIONS", default=["home"])
    create_select_context_actions: List[str] = from_settings(
        "CRUD_VIEWS_CREATE_SELECT_CONTEXT_ACTIONS", default=["home", "create_select"]
    )

    _check_messages: List[CheckMessage] = PrivateAttr(default_factory=default_list)

    @property
    def check_messages(self) -> List[CheckMessage]:

        def check_template(t):
            try:
                get_template(t)
            except TemplateDoesNotExist:
                self._check_messages.append(Error(id="E100", msg=f"template {t} not found"))

        check_template(self.extends)

        return self._check_messages

    @property
    def theme_path(self) -> str:
        return "crud_views"

    def get_js(self, path: str) -> str:
        return f"{self.theme_path}/js/{path}"

    def get_css(self, path: str) -> str:
        return f"{self.theme_path}/css/{path}"

    def javascript(self) -> dict:
        return Box(
            {
                "viewset": self.get_js("viewset.js"),
                "formset": self.get_js("formset.js"),
                "list_filter": self.get_js("list.filter.js"),
            }
        )

    @cached_property
    def css(self) -> dict:
        return Box(
            {
                "property": self.get_css("property.css"),
                "table": self.get_css("table.css"),
            }
        )

    @cached_property
    def dict(self) -> dict:
        return {
            "viewset": {
                "extends": self.extends,
                "javascript": self.javascript,
                "css": self.css,
            }
        }


crud_views_settings = CrudViewsSettings()



import pytest
from django.contrib.auth import get_user_model

from tests.lib.helper.boostrap5 import Table

User = get_user_model()


@pytest.mark.django_db
def test_author_view(client_user_author_view, cv_author, author_douglas_adams):
    client = client_user_author_view

    response = client.get("/author/")
    assert response.status_code == 200

    # list
    table = Table(response)
    row = table.rows[0]
    assert row.columns[1].text == author_douglas_adams.first_name

    # detail
    action_detail = row.get_action("detail")
    assert not action_detail.is_disabled
    response = client.get(action_detail.href)
    assert response.status_code == 200

    # create
    action_create = table.get_context_action("create")
    assert action_create.is_disabled
    response = client.get(action_create.href)
    assert response.status_code == 403

    # delete
    action_delete = row.get_action("delete")
    assert action_delete.is_disabled
    response = client.get(action_delete.href)
    assert response.status_code == 403

    # update
    action_update = row.get_action("update")
    assert action_update.is_disabled
    response = client.get(action_update.href)
    assert response.status_code == 403



from __future__ import annotations

from typing import List

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, LayoutObject, Field, BaseInput
from crud_views.lib.crispy import Column4, Column2
from django.core.exceptions import ValidationError
from django.forms import BaseForm
from django.forms.models import BaseInlineFormSet, ModelForm

from .formsets import FormSet
from .layout import FormControl


class InlineFormSet(BaseInlineFormSet):
    def __init__(
        self,
        formset: FormSet,
        parent_form: BaseForm | None = None,
        data=None,
        files=None,
        instance=None,
        save_as_new=False,
        prefix=None,
        queryset=None,
        **kwargs,
    ):
        self.formset = formset
        self.parent_form = parent_form
        super().__init__(
            data=data,
            files=files,
            instance=instance,
            save_as_new=save_as_new,
            prefix=prefix,
            queryset=queryset,
            **kwargs,
        )

    def __str__(self):
        parent = self.formset.parent.key if self.formset.parent else None
        num_forms = len(self.forms)
        return f"InlineFormSet({self.prefix},#{num_forms},m={self.form.Meta.model},p={parent})"

    __repr__ = __str__

    @property
    def helper(self) -> FormHelper:
        return self.get_helper()

    def get_helper(self) -> FormHelper:
        helper = FormHelper()
        helper.form_tag = False
        helper.disable_csrf = True
        helper.render_hidden_fields = True  # IMPORTANT: crispy needs this to render hidden fields
        helper.form_show_labels = False  # todo: get this from formset config

        fields = self.get_helper_layout_fields()
        assert isinstance(fields, list)

        hidden = self.get_helper_hidden_fields()
        assert isinstance(hidden, list)

        all_fields = fields + hidden
        helper.layout = Layout(*all_fields)
        return helper

    def get_helper_layout_fields(self) -> List[LayoutObject]:
        raise NotImplementedError()

    def get_helper_hidden_fields(self) -> List[LayoutObject]:
        fields = []
        if self.formset.can_order:
            fields.append(Field("ORDER", type="hidden"))
        if self.formset.can_delete:
            fields.append(Field("DELETE", type="hidden"))
        return fields

    @property
    def form_control(self):
        return FormControl(formset=self.formset)

    @property
    def form_control_col4(self) -> LayoutObject:
        return Column4(self.form_control, css_class="text-end")

    @property
    def form_control_col2(self) -> LayoutObject:
        return Column2(self.form_control, css_class="text-end")

    def get_form_kwargs(self, index):
        kwargs = super().get_form_kwargs(index)
        kwargs["cv_view"] = self.formset.cv_view
        return kwargs

    @property
    def has_any_form_with_data(self) -> bool:
        for data in self.cleaned_data:
            if len(data) and data.get("DELETE") is not True:
                return True
        return False

    @staticmethod
    def is_empty_form(form) -> bool:
        if form.is_valid() and not form.cleaned_data:
            return True
        else:
            # Either the form has errors (isn't valid) or
            # it doesn't have errors and contains data.
            return False

    def clean(self):
        super().clean()
        if self.parent_form:
            if self.has_any_form_with_data:
                if self.is_empty_form(self.parent_form):
                    self.parent_form.add_error(field=None, error="Child TODO requires at least one TODO set")
                    raise ValidationError("Parent form is required")


class CrispyInlineFormMixin:
    """
    This is needed for polymorphic inline formsets.
    """

    @property
    def helper(self) -> FormHelper:
        helper = FormHelper()
        helper.form_tag = False
        helper.disable_csrf = True
        helper.render_hidden_fields = True  # IMPORTANT: crispy needs this to render hidden fields
        helper.form_show_labels = True  # todo: get this from formset config
        fields = self.get_layout_fields()
        if not isinstance(fields, (list, tuple)):
            fields = [fields]
        helper.layout = Layout(*fields)
        return helper

    def get_layout_fields(self) -> LayoutObject | BaseInput | List[LayoutObject | BaseInput]:
        raise NotImplementedError


class CrispyInlineModelForm(CrispyInlineFormMixin, ModelForm):
    """
    This is needed for polymorphic inline formsets.
    """

    pass



from crispy_forms.layout import Column


class Column1(Column):
    css_class = "col-md-1"


class Column2(Column):
    css_class = "col-md-2"


class Column3(Column):
    css_class = "col-md-3"


class Column4(Column):
    css_class = "col-md-4"


class Column5(Column):
    css_class = "col-md-5"


class Column6(Column):
    css_class = "col-md-6"


class Column7(Column):
    css_class = "col-md-7"


class Column8(Column):
    css_class = "col-md-8"


class Column9(Column):
    css_class = "col-md-9"


class Column10(Column):
    css_class = "col-md-10"


class Column11(Column):
    css_class = "col-md-11"


class Column12(Column):
    css_class = "col-md-12"



from crud_views.lib.view import CrudViewPermissionRequiredMixin
from crud_views.lib.views import DetailView
from .utils import PolymorphicCrudViewMixin


class PolymorphicDetailView(PolymorphicCrudViewMixin, DetailView):
    pass


class PolymorphicDetailViewPermissionRequired(CrudViewPermissionRequiredMixin, PolymorphicDetailView):
    cv_permission = "view"



from django.apps import AppConfig


class CrudViewsPolymorphicConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "crud_views_polymorphic"
    label = "cvp"

    def ready(self):
        pass



from typing import Iterable

from django.views import generic
from django_object_detail import PropertyConfig
from django_object_detail.config import PropertyGroupConfig
from django_object_detail.views import ObjectDetailMixin

from crud_views.lib.check import Check, CheckAttribute, CheckExpression
from crud_views.lib.settings import crud_views_settings
from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin


class DetailView(ObjectDetailMixin, CrudView, generic.DetailView):
    template_name = "crud_views/view_detail.html"

    cv_key = "detail"
    cv_path = "detail"
    cv_context_actions = crud_views_settings.detail_context_actions

    # texts and labels
    cv_header_template: str | None = "crud_views/snippets/header/detail.html"
    cv_paragraph_template: str | None = "crud_views/snippets/paragraph/detail.html"
    cv_action_label_template: str | None = "crud_views/snippets/action/detail.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/detail.html"

    # icons
    cv_icon_action = "fa-regular fa-eye"

    cv_property_display: list | None = None

    @property
    def property_display(self):
        return self.cv_property_display

    @classmethod
    def checks(cls) -> Iterable[Check]:
        yield from super().checks()
        # cv_property_display must be set
        yield CheckAttribute(context=cls, id="E240", attribute="cv_property_display")
        # structural validation
        pd = cls.cv_property_display
        if pd is not None:
            yield CheckExpression(
                context=cls,
                id="E241",
                expression=isinstance(pd, list),
                msg="cv_property_display must be a list",
            )
            if isinstance(pd, list):
                for i, group in enumerate(pd):
                    if isinstance(group, PropertyGroupConfig):
                        continue
                    yield CheckExpression(
                        context=cls,
                        id="E242",
                        expression=isinstance(group, dict) and "title" in group,
                        msg=f"cv_property_display[{i}] must be a dict with a 'title' key",
                    )
                    yield CheckExpression(
                        context=cls,
                        id="E243",
                        expression=isinstance(group, dict) and "properties" in group,
                        msg=f"cv_property_display[{i}] must be a dict with a 'properties' key",
                    )
                    if isinstance(group, dict) and "properties" in group:
                        props = group["properties"]
                        yield CheckExpression(
                            context=cls,
                            id="E244",
                            expression=isinstance(props, list),
                            msg=f"cv_property_display[{i}]['properties'] must be a list",
                        )
                        if isinstance(props, list):
                            for j, prop in enumerate(props):
                                yield CheckExpression(
                                    context=cls,
                                    id="E245",
                                    expression=isinstance(prop, (str, dict, PropertyConfig)),
                                    msg=f"cv_property_display[{i}]['properties'][{j}] must be a str, dict, or PropertyConfig",
                                )


class DetailViewPermissionRequired(CrudViewPermissionRequiredMixin, DetailView):  # this file
    cv_permission = "view"



# Generated by Django 5.1.7 on 2025-04-17 18:25

import app.models.questions
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Question",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "question",
                    models.CharField(
                        max_length=100,
                        validators=[app.models.questions.validate_questionmark],
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="QuestionChoice",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order"),
                ),
                (
                    "choice",
                    models.CharField(max_length=100, validators=[app.models.questions.validate_alpha]),
                ),
                ("help_text", models.CharField(blank=True, max_length=100, null=True)),
                (
                    "question",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.question"),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="QuestionChoiceTag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order"),
                ),
                ("tag", models.CharField(max_length=100)),
                (
                    "choice",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="app.questionchoice",
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="QuestionChoiceTagAnnotation",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "order",
                    models.PositiveIntegerField(db_index=True, editable=False, verbose_name="order"),
                ),
                ("annotation", models.CharField(max_length=100)),
                (
                    "tag",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="app.questionchoicetag",
                    ),
                ),
            ],
            options={
                "ordering": ("order",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="QuestionTag",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("tag", models.CharField(max_length=100)),
                (
                    "question",
                    models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="app.question"),
                ),
            ],
        ),
    ]



from collections import UserDict
from typing import Self


class ColumnAttrs(UserDict):
    """
    Column attributes for a django-tables columns.
    ColumnAttrs can be merged with the | operator.
    """

    @classmethod
    def td_class(cls, value: str) -> Self:
        """
        Helper function to generate column attributes for the table
        """
        return cls({"td": {"class": f"{value}"}})

    @classmethod
    def th_class(cls, value: str) -> Self:
        """
        Helper function to generate column attributes for the table
        """
        return cls({"th": {"class": f"{value}"}})

    def __or__(self, other):
        assert isinstance(other, ColumnAttrs)

        def merge(dict1, dict2):
            for key, value in dict2.items():
                if key in dict1 and isinstance(dict1[key], dict) and isinstance(value, dict):
                    dict1[key] = merge(dict1[key], value)
                else:
                    if key == "class":
                        assert isinstance(value, str)
                        # check if class is already set
                        classes = [c for c in dict1.get(key, "").split(" ") if len(c)]
                        if value not in classes:
                            value = dict1.get(key, "") + " " + value
                            dict1[key] = value
                    else:
                        dict1[key] = value
            return dict1

        merged: dict = merge(dict(self), dict(other))
        return ColumnAttrs(merged)


class ColAttr:
    # td width
    ID: ColumnAttrs = ColumnAttrs.td_class("cv-col-id")
    w5: ColumnAttrs = ColumnAttrs.td_class("cv-col-5")
    w10: ColumnAttrs = ColumnAttrs.td_class("cv-col-10")
    w15: ColumnAttrs = ColumnAttrs.td_class("cv-col-15")
    w20: ColumnAttrs = ColumnAttrs.td_class("cv-col-20")
    w25: ColumnAttrs = ColumnAttrs.td_class("cv-col-25")
    w30: ColumnAttrs = ColumnAttrs.td_class("cv-col-30")
    w35: ColumnAttrs = ColumnAttrs.td_class("cv-col-35")
    w40: ColumnAttrs = ColumnAttrs.td_class("cv-col-40")
    w45: ColumnAttrs = ColumnAttrs.td_class("cv-col-45")
    w50: ColumnAttrs = ColumnAttrs.td_class("cv-col-50")
    w55: ColumnAttrs = ColumnAttrs.td_class("cv-col-55")
    w60: ColumnAttrs = ColumnAttrs.td_class("cv-col-60")
    w65: ColumnAttrs = ColumnAttrs.td_class("cv-col-65")
    w70: ColumnAttrs = ColumnAttrs.td_class("cv-col-70")
    w75: ColumnAttrs = ColumnAttrs.td_class("cv-col-75")
    w80: ColumnAttrs = ColumnAttrs.td_class("cv-col-80")
    w85: ColumnAttrs = ColumnAttrs.td_class("cv-col-85")
    w90: ColumnAttrs = ColumnAttrs.td_class("cv-col-90")
    w95: ColumnAttrs = ColumnAttrs.td_class("cv-col-95")
    w100: ColumnAttrs = ColumnAttrs.td_class("cv-col-100")

    # extra
    action: ColumnAttrs = ColumnAttrs.th_class("d-flex justify-content-end") | ColumnAttrs.td_class(
        "d-flex justify-content-end"
    )



class PrimaryKeys:
    """
    Predefined primary key regexes as strings
    """

    INT: str = r"\d+"
    HEX: str = r"[0-9a-z]+"
    UUID: str = r"[0-9a-f]{8}-[0-9a-f]{4}-4[0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}"
    KEY: str = "[a-z_]+"
    STR: str = r"[A-Za-z0-9_\-]+"


def get_path_pk(name: str, reg: str) -> str:
    return rf"(?P<{name}>{reg})"


def get_path(name: str, reg: str) -> str:
    return rf"(?P<{name}>{reg})"



from functools import cached_property
from typing import Dict, Any, List

from django.contrib.contenttypes.models import ContentType
from django.db.models import Choices
from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from ..models import WorkflowInfo
from .enums import BadgeEnum, WorkflowComment

TState = str


class WorkflowModelMixin:
    """
    Mixin for workflow models with a django-fsm state field
    """

    STATE_CHOICES: Choices = None  # i.e. models.TextChoice
    STATE_BADGES: Dict[TState, BadgeEnum] | None = None
    STATE_BADGE_DEFAULT: BadgeEnum = BadgeEnum.INFO
    COMMENT_DEFAULT: WorkflowComment = WorkflowComment.NONE

    def state_increment(self):
        self.state_version += 1  # noqa

    def get_state_name(self, state) -> str:
        assert self.STATE_CHOICES, "STATE_CHOICES must be defined"  # noqa
        if not state:
            return ""
        return dict(self.STATE_CHOICES.choices)[state]  # noqa

    @property
    def state_name(self) -> str:
        assert self.state, "FSM state column missing"  # noqa
        return self.get_state_name(self.state)  # noqa

    def get_state_badge(self, state) -> str:
        assert self.STATE_BADGES is not None, "STATE_BADGES must be defined"  # noqa
        name = self.get_state_name(state)
        klass = self.STATE_BADGES.get(state, self.STATE_BADGE_DEFAULT)  # noqa
        html = render_to_string("crud_views_workflow/badge.html", {"state": state, "name": name, "class": klass})
        return mark_safe(html)

    @property
    def state_badge(self) -> str:
        return self.get_state_badge(self.state)  # noqa

    @staticmethod
    def workflow_comment_kwargs(comment: str | None = None) -> Dict:
        return dict(comment=comment) if comment else dict()

    def get_workflow_info_queryset(self):
        return WorkflowInfo.objects.filter(
            workflow_object_pk=str(self.pk),
            workflow_object_content_type=ContentType.objects.get_for_model(self),
        ).order_by("timestamp")  # noqa

    @property
    def workflow_data(self) -> List[Dict]:
        items = []
        queryset = self.get_workflow_info_queryset()
        for info in queryset:
            items.append(
                dict(
                    timestamp=info.timestamp,
                    user=info.user,
                    state_old=info.state_old,
                    state_new=info.state_new,
                    state_old_badge=self.get_state_badge(info.state_old),
                    state_new_badge=self.get_state_badge(info.state_new),
                    transition=info.transition,
                    transition_label=self.workflow_get_transition_label(info.transition),
                    comment=info.comment,
                )
            )
        return items

    def workflow_get_possible_transitions(self, user) -> List[tuple[Any, Any, Any]]:
        actions = []
        for transition in self.get_available_user_state_transitions(user):  # noqa
            label = transition.custom.get("label", transition.name)
            comment = transition.custom.get("comment", self.COMMENT_DEFAULT)
            actions.append((transition.name, label, comment))
        return actions

    def workflow_has_any_possible_transition(self, user) -> bool:
        return len(self.workflow_get_possible_transitions(user)) > 0

    def workflow_has_transition(self, user, transition) -> bool:
        for v, x, y in self.workflow_get_possible_transitions(user):
            if v == transition:
                return True
        return False

    @cached_property
    def workflow_get_transition_label_map(self) -> Dict[str, str]:
        return {t.name: t.custom.get("label", t.name) for t in self.get_all_state_transitions()}

    def workflow_get_transition_label(self, name) -> str:
        return self.workflow_get_transition_label_map[name]

    def workflow_get_form_kwargs(self, user):
        kwargs, choices, _initial, transition_comments = dict(), [], [], dict()
        for value, label, comment in self.workflow_get_possible_transitions(user):
            choices.append((value, label))
            transition_comments[value] = comment
        kwargs["choices"] = choices
        kwargs["transition_comments"] = transition_comments
        return kwargs



from crud_views.lib.settings import crud_views_settings
from crud_views.lib.view import CrudView, CrudViewPermissionRequiredMixin
from django.views import generic

from .mixins import CrudViewProcessFormMixin


class CreateView(CrudViewProcessFormMixin, CrudView, generic.CreateView):
    template_name = "crud_views/view_create.html"

    cv_key = "create"
    cv_object = False
    cv_path = "create"
    cv_success_key = "list"
    cv_context_actions = crud_views_settings.create_context_actions

    # texts and labels
    cv_header_template: str | None = "crud_views/snippets/header/create.html"
    cv_paragraph_template: str | None = "crud_views/snippets/paragraph/create.html"
    cv_action_label_template: str | None = "crud_views/snippets/action/create.html"
    cv_action_short_label_template: str | None = "crud_views/snippets/action_short/create.html"

    # icons
    cv_icon_action = "fa-regular fa-square-plus"

    # messages
    cv_message_template: str | None = "crud_views/snippets/message/create.html"

    def cv_form_valid(self, context: dict):
        """
        Handle valid form: save the form instance
        """
        self.object = context["form"].save()
        context["object"] = self.object


class CreateViewPermissionRequired(CrudViewPermissionRequiredMixin, CreateView):  # this file
    cv_permission = "add"


class CreateViewParentMixin:
    def cv_form_valid(self, context: dict):
        """
        Handle valid form
        """
        assert self.cv_viewset.has_parent, "this ViewSet has no parent"

        parent_object = self.cv_get_parent_object()

        if self.cv_viewset.parent.many_to_many_through_attribute:
            # this saves the object first
            super().cv_form_valid(context)

            # get m2m attribute
            m2m = getattr(parent_object, self.cv_viewset.parent.many_to_many_through_attribute)

            # get through defaults
            through_defaults = self.cv_parent_many_to_many_through_defaults(self.object, parent_object, m2m)

            # add the object to the m2m attribute
            m2m.add(self.object, through_defaults=through_defaults)
        else:
            # before saving, add the parent model to the form instance
            setattr(context["form"].instance, self.cv_get_parent_object_attribute(), parent_object)

            # now save the form
            super().cv_form_valid(context)

    def cv_parent_many_to_many_through_defaults(self, instance, parent_instance, m2m) -> dict:
        return dict()



import pytest
from django.test.client import Client

from crud_views.lib.viewset import ViewSet
from tests.lib.helper.boostrap5 import Table
from tests.test1.app.models import Publisher


@pytest.mark.django_db
def test_viewset_has_views(cv_publisher: ViewSet):
    expected_views = ["list", "detail", "create", "update", "delete"]
    for view_name in expected_views:
        assert cv_publisher.has_view(view_name), f"ViewSet missing view: {view_name}"


@pytest.mark.django_db
def test_viewset_url_patterns(cv_publisher: ViewSet):
    patterns = cv_publisher.urlpatterns
    assert len(patterns) >= 5

    # Verify INT pk regex is used (not UUID)
    routes = [p.pattern.regex.pattern for p in patterns]
    has_int_pk = any(r"\d+" in route for route in routes)
    assert has_int_pk, f"Expected INT pk regex in patterns, got: {routes}"

    has_uuid = any("4[0-9a-f]" in route for route in routes)
    assert not has_uuid, "Should not have UUID pk regex"


@pytest.mark.django_db
def test_viewset_router_names(cv_publisher: ViewSet):
    assert cv_publisher.get_router_name("list") == "publisher-list"
    assert cv_publisher.get_router_name("detail") == "publisher-detail"
    assert cv_publisher.get_router_name("create") == "publisher-create"
    assert cv_publisher.get_router_name("update") == "publisher-update"
    assert cv_publisher.get_router_name("delete") == "publisher-delete"


@pytest.mark.django_db
def test_create_publisher(client_user_publisher_add: Client, cv_publisher):
    response = client_user_publisher_add.get("/publisher/create/")
    assert response.status_code == 200

    response = client_user_publisher_add.post(
        "/publisher/create/",
        {
            "name": "Penguin Random House",
        },
    )
    assert response.status_code == 302
    assert response.url == "/publisher/"

    assert Publisher.objects.filter(name="Penguin Random House").exists()


@pytest.mark.django_db
def test_update_publisher(client_user_publisher_change: Client, cv_publisher, publisher_penguin):
    pk = publisher_penguin.pk

    response = client_user_publisher_change.get(f"/publisher/{pk}/update/")
    assert response.status_code == 200

    response = client_user_publisher_change.post(
        f"/publisher/{pk}/update/",
        {
            "name": "Penguin Classics",
        },
    )
    assert response.status_code == 302
    assert response.url == "/publisher/"

    publisher_penguin.refresh_from_db()
    assert publisher_penguin.name == "Penguin Classics"


@pytest.mark.django_db
def test_delete_publisher(client_user_publisher_delete: Client, cv_publisher, publisher_penguin):
    pk = publisher_penguin.pk

    response = client_user_publisher_delete.get(f"/publisher/{pk}/delete/")
    assert response.status_code == 200

    response = client_user_publisher_delete.post(f"/publisher/{pk}/delete/", {"confirm": True})
    assert response.status_code == 302
    assert response.url == "/publisher/"

    assert not Publisher.objects.filter(pk=pk).exists()


@pytest.mark.django_db
def test_detail_publisher(client_user_publisher_view: Client, cv_publisher, publisher_penguin):
    pk = publisher_penguin.pk

    response = client_user_publisher_view.get(f"/publisher/{pk}/detail/")
    assert response.status_code == 200

    content = response.content.decode("utf-8")
    assert "Penguin" in content


@pytest.mark.django_db
def test_list_publisher(client_user_publisher_view: Client, cv_publisher, publisher_penguin):
    response = client_user_publisher_view.get("/publisher/")
    assert response.status_code == 200

    table = Table(response)
    assert len(table.rows) == 1
    assert table.rows[0].columns[1].text == "Penguin"



from crud_views.lib.view import CrudViewPermissionRequiredMixin
from crud_views.lib.views import UpdateView
from .utils import PolymorphicCrudViewMixin


class PolymorphicUpdateView(PolymorphicCrudViewMixin, UpdateView):
    pass


class PolymorphicUpdateViewPermissionRequired(CrudViewPermissionRequiredMixin, PolymorphicUpdateView):
    cv_permission = "change"



import django_tables2 as tables
from django_tables2 import Column

from crud_views.lib.exceptions import ViewSetKeyFoundError, ignore_exception
from crud_views.lib.table.attrs import ColAttr

from crud_views.lib.viewset import ViewSet


class ViewAwareColumnMixin:
    pass


class ActionColumn(tables.TemplateColumn):
    def __init__(self, **extra):
        if "orderable" not in extra:
            extra["orderable"] = False
        if "template_name" not in extra:
            extra["template_name"] = "crud_views/columns/actions.html"
        if "attrs" not in extra:
            extra["attrs"] = ColAttr.action
        super().__init__(**extra)

    def render(self, record, table, value, bound_column, **kwargs):
        self.extra_context["view"] = table.view
        return super().render(record, table, value, bound_column, **kwargs)


class LinkChildColumn(tables.TemplateColumn):
    def __init__(self, name: str, key: str = "list", **extra):
        self.name = name
        self.key = key
        if "orderable" not in extra:
            extra["orderable"] = False
        if "template_code" not in extra and "template_name" not in extra:
            extra["template_name"] = "crud_views/columns/child.html"
        super().__init__(**extra)

    def render(self, record, table, value, bound_column, **kwargs):

        viewset = ViewSet.get_viewset(self.name)
        data = viewset.get_meta(table.view.cv_get_view_context())
        data.update({"url": table.view.cv_get_child_url(self.name, self.key, record)})
        self.extra_context.update(data)
        return super().render(record, table, value, bound_column, **kwargs)


class LinkDetailColumnMixin:
    @ignore_exception(ViewSetKeyFoundError)
    def get_url(self, table, record, **kwargs):
        return table.view.cv_get_url("detail", obj=record)


class LinkDetailColumn(LinkDetailColumnMixin, Column):
    pass


class UUIDColumn(tables.TemplateColumn):
    def __init__(self, template_name="crud_views/columns/uuid.html", **extra):
        if "orderable" not in extra:
            extra["orderable"] = True
        super().__init__(template_name=template_name, **extra)

    def render(self, record, table, value, bound_column, **kwargs):
        self.extra_context["view"] = table.view
        self.extra_context["uuid_short"] = str(value).split("-")[0]
        return super().render(record, table, value, bound_column, **kwargs)


class UUIDLinkDetailColumn(LinkDetailColumnMixin, UUIDColumn):
    pass


class NaturalTimeColumn(tables.TemplateColumn):
    def __init__(self, template_code=None, template_name="crud_views/columns/naturaltime.html", **extra):
        from .table import Table

        if "attrs" not in extra:
            extra["attrs"] = Table.ca.w10

        super().__init__(template_code=template_code, template_name=template_name, **extra)


class NaturalDayColumn(tables.TemplateColumn):
    def __init__(self, template_code=None, template_name="crud_views/columns/naturalday.html", **extra):
        from .table import Table

        if "attrs" not in extra:
            extra["attrs"] = Table.ca.w10

        super().__init__(template_code=template_code, template_name=template_name, **extra)



import django_tables2 as tables
from crispy_forms.layout import Row
from django.utils.translation import gettext as _

from app.models.poly import Parent
from crud_views.lib.crispy import CrispyModelViewMixin, CrispyDeleteForm, CrispyModelForm, Column4
from crud_views.lib.table import Table, UUIDLinkDetailColumn
from crud_views.lib.views import (
    ListViewTableMixin,
    ListViewPermissionRequired,
    RedirectChildView,
    CreateViewPermissionRequired,
    UpdateViewPermissionRequired,
    DeleteViewPermissionRequired,
)
from crud_views.lib.viewset import ViewSet

cv_poly_parent_formset = ViewSet(model=Parent, name="parent", icon_header="fa-solid fa-user-group")


class ParentForm(CrispyModelForm):
    submit_label = "Create"

    class Meta:
        model = Parent
        fields = ["name"]

    def get_layout_fields(self):
        return Row(Column4("name"))


class ParentTable(Table):
    id = UUIDLinkDetailColumn(attrs=Table.ca.ID)
    name = tables.Column(attrs=Table.ca.w80)


class ParentListView(ListViewTableMixin, ListViewPermissionRequired):
    model = Parent

    cv_viewset = cv_poly_parent_formset
    cv_list_actions = ["redirect_child", "update", "delete"]

    table_class = ParentTable


class ParentCreateView(CrispyModelViewMixin, CreateViewPermissionRequired):
    model = Parent
    form_class = ParentForm
    cv_viewset = cv_poly_parent_formset


class ParentUpdateView(CrispyModelViewMixin, UpdateViewPermissionRequired):
    model = Parent
    form_class = ParentForm
    cv_viewset = cv_poly_parent_formset


class ParentDeleteView(CrispyModelViewMixin, DeleteViewPermissionRequired):
    model = Parent
    form_class = CrispyDeleteForm
    cv_viewset = cv_poly_parent_formset


class RedirectPolyView(RedirectChildView):
    cv_action_label = _("Manage Poly")
    cv_redirect = "poly"
    cv_redirect_key = "list"
    cv_icon_action = "fa-solid fa-sun"

    cv_viewset = cv_poly_parent_formset



from django import template
from django.contrib.auth import get_user_model
from django.template.loader import render_to_string

from crud_views.lib.settings import crud_views_settings
from crud_views.lib.exceptions import ViewSetKeyFoundError, ignore_exception
from crud_views.lib.view import CrudView

User = get_user_model()

register = template.Library()


@register.inclusion_tag(f"{crud_views_settings.theme_path}/shared/css.html", takes_context=True)
def cv_css(context):
    return {"css": crud_views_settings.css}


@register.inclusion_tag(f"{crud_views_settings.theme_path}/shared/js.html", takes_context=True)
def cv_js(context):
    return {"js": crud_views_settings.javascript}


@register.inclusion_tag(f"{crud_views_settings.theme_path}/shared/csrftoken.html", takes_context=True)
def cv_csrf_token(context):
    return {}


def cv_get_view(context) -> CrudView:
    view: CrudView = context["view"]
    assert isinstance(view, CrudView), f"view {view} is not ViewSetAware"
    return view


def cv_get_context(context, key, obj=None) -> dict:
    view: CrudView = cv_get_view(context)
    context = view.cv_get_context(key, obj=obj, user=context["request"].user, request=context["request"])
    return context


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/const.js.html", takes_context=True)
def cv_const_js(context):
    request = context["request"]
    return {
        "request_path": request.path,
        "request_query_string": request.META.get("QUERY_STRING", ""),
    }


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/list_action.html", takes_context=True)
def cv_list_action(context, key, obj=None):
    return cv_get_context(context=context, key=key, obj=obj)


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/list_action_form.html", takes_context=True)
def cv_list_action_form(context, key, obj=None):
    return cv_get_context(context=context, key=key, obj=obj)


@register.simple_tag(takes_context=True)
@ignore_exception(ViewSetKeyFoundError, default_value="")
def cv_context_action(context, key, obj=None):
    ctx = cv_get_context(context=context, key=key, obj=obj)
    template = ctx.get("cv_template", f"{crud_views_settings.theme_path}/tags/context_action.html")
    return render_to_string(template, context=ctx, request=context["request"])


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/context_actions.html", takes_context=True)
def cv_context_actions(context, obj=None):
    view: CrudView = cv_get_view(context)
    return {"view": view, "request": context["request"], "object": obj}


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/button_cancel.html", takes_context=True)
def cv_cancel_button(context, obj=None):
    view: CrudView = cv_get_view(context)
    context = view.get_cancel_button_context(obj=obj, user=context["request"].user, request=context["request"])
    return context


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/button_submit.html", takes_context=True)
def cv_submit_button(context, obj=None):
    cv_get_view(context)
    return {}


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/button_delete.html", takes_context=True)
def cv_delete_button(context, obj=None):
    cv_get_view(context)
    return {}


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/icon.html", takes_context=True)
def cv_header_icon(context):
    view: CrudView = cv_get_view(context)
    icon = view.cv_get_header_icon()
    return {"icon": icon}


@register.inclusion_tag(f"{crud_views_settings.theme_path}/tags/icon.html", takes_context=True)
def cv_filter_icon(context):
    view: CrudView = cv_get_view(context)
    icon = view.cv_get_filter_icon()  # noqa
    return {"icon": icon}


@register.simple_tag(takes_context=True)
def cv_filter_header(context):
    view: CrudView = cv_get_view(context)
    return view.cv_filter_header  # noqa


@register.simple_tag(takes_context=True)
def cv_header(context):
    view: CrudView = cv_get_view(context)
    return view.cv_header


@register.simple_tag(takes_context=True)
def cv_paragraph(context):
    view: CrudView = cv_get_view(context)
    return view.cv_paragraph


@register.simple_tag(takes_context=True)
def cv_render_form(context):
    return render_to_string(f"{crud_views_settings.theme_path}/tags/form.html", context.flatten())


@register.simple_tag(takes_context=True)
def cv_render_filter(context):
    return render_to_string(f"{crud_views_settings.theme_path}/tags/list_filter.html", context.flatten())


@register.filter
def cv_is_false(arg):
    return arg is False


@register.filter
def cv_is_true(arg):
    return arg is True



# Generated by Django 4.2 on 2026-02-20 08:08

import crud_views_workflow.lib.mixins
from django.db import migrations, models
import django_fsm


class Migration(migrations.Migration):
    dependencies = [
        ("app", "0003_alter_polyanswer_options_and_more"),
    ]

    operations = [
        migrations.CreateModel(
            name="Campaign",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("name", models.CharField(max_length=128)),
                (
                    "state",
                    django_fsm.FSMField(
                        choices=[
                            ("new", "New"),
                            ("active", "Created"),
                            ("success", "Error"),
                            ("canceled", "Assigned"),
                        ],
                        default="new",
                        max_length=50,
                    ),
                ),
            ],
            bases=(crud_views_workflow.lib.mixins.WorkflowModelMixin, models.Model),
        ),
    ]
